#!/usr/bin/env python3
# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
import ipaddress
import sys
import time
from collections import OrderedDict
import re
from textwrap import dedent
from typing import List, Dict, Optional, Union, cast, Iterable

import yaml
import ruamel.yaml
from deepdiff import DeepDiff  # type: ignore[import-untyped]

from ordered_set import OrderedSet

from kubemarine import (
    packages as pckgs, system, selinux, etcd, thirdparties, apparmor, kubernetes, sysctl, audit,
    plugins, modprobe, admission
)
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.group import NodeGroup, CollectorCallback, GroupResultException
from kubemarine.cri import containerd
from kubemarine.kubernetes import components
from kubemarine.plugins import calico, builtin, manifest
from kubemarine.procedures import check_iaas
from kubemarine.core import flow, static, utils
from kubemarine.testsuite import TestSuite, TestCase, TestFailure, TestWarn
from kubemarine.kubernetes.daemonset import DaemonSet
from kubemarine.kubernetes.deployment import Deployment
from kubemarine.kubernetes.object import KubernetesObject
from kubemarine.coredns import generate_configmap


def services_status(cluster: KubernetesCluster, service_type: str) -> None:
    with TestCase(cluster, '201', "Services", "%s Status" % service_type.capitalize(),
                  default_results='active (running)'):
        group = cluster.nodes['all']
        if service_type == 'haproxy':
            group = cluster.make_group_from_roles(['balancer'])
        elif service_type == 'keepalived':
            group = cluster.make_group_from_roles(['keepalived'])
        elif service_type in ('containerd', 'kubelet'):
            group = cluster.make_group_from_roles(['control-plane', 'worker'])

        if group.is_empty():
            raise TestWarn("No nodes to check service status",
                           hint="The node group to check the service is empty. Check skipped.")

        collector = CollectorCallback(cluster)
        with group.new_executor() as exe:
            for node in exe.group.get_ordered_members_list():
                service_name = service_type
                if service_type != 'kubelet':
                    service_name = cluster.get_package_association_for_node(
                        node.get_host(), service_type, 'service_name')
                system.service_status(node, name=service_name, callback=collector)

        cluster.log.verbose(collector.result)

        status_regexp = re.compile(r"Active:\s([a-z\s()]*)(\ssince|$)", re.M)

        statuses = []
        failed = False
        for host, node_result in collector.result.items():
            if node_result.return_code == 4:
                statuses.append('service is missing')
                failed = True
                cluster.log.debug('%s is not presented on host %s, skipped'
                                  % (service_type.capitalize(), host))
                continue
            matches = re.findall(status_regexp, node_result.stdout)
            if matches:
                status = matches[0][0].strip()
                cluster.log.debug(
                    '%s status is \"%s\" at host %s' % (service_type.capitalize(), status, host))
                if status != 'active (running)':
                    statuses.append(status)
                    failed = True
            elif node_result.return_code != 0:
                failed = True
                cluster.log.error('%s status has bad exit code \"%s\" at host %s'
                                  % (service_type.capitalize(), node_result.return_code, host))
            else:
                raise Exception('Failed to detect status for \"%s\"' % host)

        statuses = list(set(statuses))

        if failed:
            raise TestFailure("Bad status detected: %s" % ', '.join(statuses),
                              hint="Fix the service to be enabled and has running status.")


def _check_same_os(cluster: KubernetesCluster) -> None:
    os_ids = cluster.get_os_identifiers()
    different_os = set(os_ids.values())
    if len(different_os) > 1:
        cluster.log.warning(
            f"Nodes have different OS families or versions, packages versions cannot be checked. "
            f"List of (OS family, version): {list(different_os)}")
        raise TestFailure(f"Nodes have different OS families or versions")


def recommended_system_package_versions(cluster: KubernetesCluster, pckg_alias: str) -> None:
    """
    Function that checks if defined package are compatible with the configured k8s version and OS.
    Raise Warn if unable to detect the OS family, configured not recommended k8s version or
    if configured not recommended system packages versions.
    """
    os_family = cluster.get_os_family()
    if os_family not in pckgs.get_associations_os_family_keys():
        raise TestFailure("OS is unknown or multiple OS present")

    recommended_packages = list(static.GLOBALS['packages'][os_family])
    if pckg_alias not in recommended_packages:
        raise TestWarn(f"Package {pckg_alias} doesn't have recommended version")

    k8s_version = cluster.inventory['services']['kubeadm']['kubernetesVersion']
    expected_system_packages = pckgs.get_default_package_names(os_family, pckg_alias, k8s_version)

    good_results = set()
    bad_results = []
    actual_packages = cluster.get_package_association(pckg_alias, "package_name")
    if not isinstance(actual_packages, list):
        actual_packages = [actual_packages]
    for expected_pckg in expected_system_packages:
        expected_pckg_name = pckgs.get_package_name(os_family, expected_pckg)
        for actual_pckg in actual_packages:
            actual_pckg_name = pckgs.get_package_name(os_family, actual_pckg)
            if expected_pckg_name == actual_pckg_name:
                if actual_pckg.startswith(expected_pckg.replace("*", "")):
                    good_results.add(actual_pckg)
                else:
                    cluster.log.debug(f"Package {actual_pckg} is not recommended, recommended is {expected_pckg}")
                    bad_results.append(actual_pckg)

                break
        else:
            cluster.log.debug(f"Package {expected_pckg_name} is not found in inventory")
            bad_results.append(expected_pckg_name)

    if bad_results:
        raise TestWarn(f"Detected not recommended packages versions: {bad_results}")
    cluster.log.debug(f"Detected packages with recommended versions: {good_results}")


def system_packages_versions(cluster: KubernetesCluster, pckg_alias: str) -> None:
    """
    Verifies that system packages are installed on required nodes and have equal versions.
    Failure is shown if check is not successful.
    :param cluster: main cluster object.
    :param pckg_alias: system package alias to retrieve "package_name" association.
    """
    with TestCase(cluster, '205', "Services", f"{pckg_alias} version") as tc:
        _check_same_os(cluster)
        hosts_to_packages = pckgs.get_association_hosts_to_packages(cluster.nodes['all'], cluster.inventory, pckg_alias)
        if not hosts_to_packages:
            raise TestWarn(f"No nodes to check {pckg_alias!r} version")

        os_family = cluster.get_os_family()
        require_same_version = []
        if pckgs.cache_versions_enabled(cluster, os_family, pckg_alias):
            require_same_version = pckgs.get_association_packages(cluster, os_family, pckg_alias)

        check_packages_versions(cluster, tc, hosts_to_packages, require_same_version, raise_successful=False)

        if pckg_alias in ["haproxy", "keepalived", "containerd"]:
            recommended_system_package_versions(cluster, pckg_alias)
        tc.success("all packages have correct versions")


def mandatory_packages_versions(cluster: KubernetesCluster) -> None:
    """
    Verifies that mandatory packages are installed on required nodes and have equal versions.
    Failure is shown if check is not successful.
    :param cluster: main cluster object.
    """
    with TestCase(cluster, '205', "Services", "Mandatory package versions") as tc:
        _check_same_os(cluster)
        hosts_to_packages: Dict[str, List[str]] = {}
        group = cluster.nodes['all']
        os_family = cluster.get_os_family()
        require_same_version = OrderedSet[str]()
        for package in cluster.inventory["services"]["packages"]['mandatory'].keys():
            packages = pckgs.get_association_hosts_to_packages(group, cluster.inventory, package)

            for host, packages_list in packages.items():
                hosts_to_packages.setdefault(host, []).extend(packages_list)

            if pckgs.cache_versions_enabled(cluster, os_family, package):
                require_same_version.update(pckgs.get_association_packages(cluster, os_family, package))

        if not hosts_to_packages:
            raise TestWarn(f"No mandatory packages to check")

        return check_packages_versions(cluster, tc, hosts_to_packages, require_same_version)


def generic_packages_versions(cluster: KubernetesCluster) -> None:
    """
    Verifies that user-provided packages are installed on required nodes and have equal versions.
    Warning is shown if check is not successful.
    """
    with TestCase(cluster, '206', "Services", f"Generic packages version") as tc:
        _check_same_os(cluster)
        packages = cluster.inventory['services']['packages'].get('install', {}).get('include', [])
        hosts_to_packages = {host: packages for host in cluster.nodes['all'].get_hosts()}
        return check_packages_versions(cluster, tc, hosts_to_packages, packages, warn_on_bad_result=True)


def check_packages_versions(cluster: KubernetesCluster, tc: TestCase, hosts_to_packages: Dict[str, List[str]],
                            packages_require_same_version: Iterable[str],
                            warn_on_bad_result: bool = False, raise_successful: bool = True) -> None:
    """
    Verifies that all packages are installed on required nodes and have equal versions
    :param cluster: main cluster object
    :param tc: current test case object
    :param hosts_to_packages: hosts where to check packages
    :param packages_require_same_version: Fail or Warn for the specified set of packages
                                          if different versions are detected.
    :param warn_on_bad_result: if true then bad results raises Warning instead of Failure. Default False.
    :param raise_successful: if true, successful test result will be risen. Default True.
    """
    bad_results = OrderedSet[str]()
    warn_results = OrderedSet[str]()
    good_results = []

    packages_map = pckgs.detect_installed_packages_version_hosts(cluster, hosts_to_packages)
    for package, version_map in packages_map.items():
        installed_versions = [version for version in version_map if "not installed" not in version]
        if len(installed_versions) != 1:
            cluster.log.debug(f"Package {package} has different versions:")
            cluster.log.debug(installed_versions)
            if package in packages_require_same_version:
                bad_results.append(package)
            else:
                warn_results.append(package)

        for version in version_map.keys():
            if "not installed" in version:
                cluster.log.debug(f"Package {package} is not installed on some nodes:")
                cluster.log.debug(version_map[version])
                bad_results.append(package)
            else:
                package_name = package.replace("*", "")
                if version.startswith(package_name):
                    good_results.append(version)
                else:
                    cluster.log.debug(f"Actual version: {version} doesn't match expected package version: {package} on "
                                      f"nodes: {version_map[version]}")
                    bad_results.append(version)

    if bad_results and not warn_on_bad_result:
        hint_message = f'Check the presence and correctness of the version of the following packages on the ' \
                       f'system: {list(bad_results)}\n'
        raise TestFailure("detected incorrect packages versions", hint=hint_message)
    if warn_results or (warn_on_bad_result and bad_results):
        hint_message = f'Check the presence and correctness of the version of the following packages on the ' \
                       f'system: {list(bad_results | warn_results)}\n'
        raise TestWarn("detected incorrect packages versions", hint=hint_message)
    cluster.log.debug(f"installed packages: {good_results}")
    if raise_successful:
        tc.success("all packages have correct versions")


def get_nodes_description(cluster: KubernetesCluster) -> dict:
    return kubernetes.get_nodes_description(cluster)


def kubelet_version(cluster: KubernetesCluster) -> None:
    with TestCase(cluster, '203', "Services", "Kubelet Version",
                  default_results=cluster.inventory['services']['kubeadm']['kubernetesVersion']):
        nodes_description = get_nodes_description(cluster)
        bad_versions = []
        for node_description in nodes_description['items']:
            node_name = node_description['metadata']['name']
            kubelet_version = node_description['status']['nodeInfo']['kubeletVersion']
            cluster.log.debug("Node \"%s\" running kubelet %s" % (node_name, kubelet_version))
            if kubelet_version != cluster.inventory['services']['kubeadm']['kubernetesVersion']:
                bad_versions.append(kubelet_version)
        bad_versions = list(set(bad_versions))
        if bad_versions:
            raise TestFailure("Invalid version detected: %s" % ', '.join(bad_versions),
                              hint="All nodes must have the same correct Kubelet version \"%s\". Remove nodes with the "
                                   "incorrect version from the cluster and reinstall them to the corresponding "
                                   "versions." % cluster.inventory['services']['kubeadm']['kubernetesVersion'])


def kubelet_config(cluster: KubernetesCluster) -> None:
    """
    This test checks the consistency of the /var/lib/kubelet/config.yaml configuration
    with `kubelet-config` ConfigMap and with the inventory.
    """
    with TestCase(cluster, '233', "Services", "Kubelet Configuration") as tc:
        messages = []

        cluster.log.debug("Checking configuration consistency with kubelet-config ConfigMap")
        failed_nodes = _compare_kubelet_config(cluster, with_inventory=False)
        if failed_nodes:
            messages.append(f"/var/lib/kubelet/config.yaml is not consistent with kubelet-config ConfigMap "
                            f"on nodes {', '.join(failed_nodes)}")

        cluster.log.debug("Checking configuration consistency with patches from inventory")
        failed_nodes = _compare_kubelet_config(cluster, with_inventory=True)
        if failed_nodes:
            messages.append(f"/var/lib/kubelet/config.yaml is not consistent with patches from inventory "
                            f"on nodes {', '.join(failed_nodes)}")

        cluster.log.debug("Checking kubelet-config ConfigMap consistency with services.kubeadm_kubelet section")
        diff = components.compare_configmap(cluster, 'kubelet-config')
        if diff is not None:
            msg = "kubelet-config ConfigMap is not consistent with services.kubeadm_kubelet section"
            messages.append(msg)
            cluster.log.debug(msg)
            cluster.log.debug(diff + '\n')

        if not messages:
            tc.success(results='valid')
        else:
            messages.append(
                "Check DEBUG logs for details. To have the check passed, "
                "you may need to call `kubemarine reconfigure --tasks deploy.kubernetes.reconfigure`, "
                "or change the inventory file accordingly."
            )
            raise TestFailure('invalid', hint=yaml.safe_dump(messages))


def _compare_kubelet_config(cluster: KubernetesCluster, *, with_inventory: bool) -> List[str]:
    config_diffs = components.compare_kubelet_config(cluster, with_inventory=with_inventory)
    failed_nodes = OrderedSet[str]()
    for host, diff in config_diffs.items():
        if diff is None:
            continue

        node_name = cluster.get_node_name(host)
        cluster.log.debug(f"/var/lib/kubelet/config.yaml is not actual on node {node_name}")
        if failed_nodes:
            cluster.log.debug('<truncated, enable verbose logs for details>')
            cluster.log.verbose(diff + '\n')
        else:
            cluster.log.debug(diff + '\n')

        failed_nodes.add(node_name)

    return list(failed_nodes)


def kube_proxy_configmap(cluster: KubernetesCluster) -> None:
    """
    This test checks the consistency of the `kube-proxy` ConfigMap with the inventory.
    """
    with TestCase(cluster, '234', "Services", "kube-proxy Configuration") as tc:
        messages = []

        cluster.log.debug("Checking kube-proxy ConfigMap consistency with services.kubeadm_kube-proxy section")
        diff = components.compare_configmap(cluster, 'kube-proxy')
        if diff is not None:
            msg = "kube-proxy ConfigMap is not consistent with services.kubeadm_kube-proxy section"
            messages.append(msg)
            cluster.log.debug(msg)
            cluster.log.debug(diff + '\n')

        if not messages:
            tc.success(results='valid')
        else:
            messages.append(
                "Check DEBUG logs for details. To have the check passed, "
                "you may need to call `kubemarine reconfigure --tasks deploy.kubernetes.reconfigure`, "
                "or change the inventory file accordingly."
            )
            raise TestFailure('invalid', hint=yaml.safe_dump(messages))


def thirdparties_hashes(cluster: KubernetesCluster) -> None:
    """
    Task which is used to verify configured thirdparties hashes agains actual hashes on nodes.
    If thirdparty is an archive, then archive files hashes are also verified.
    If hash is not specified, then thirdparty is skipped.
    If there is no thirdparties with hashes, then warning is shown.
    """
    with TestCase(cluster, '212', "Thirdparties", "Hashes") as tc:
        successful = []
        warnings = []
        broken = []

        #Create tmp dir for loading thirdparty without default sha
        first_control_plane = cluster.nodes['control-plane'].get_first_member()
        first_control_plane_host = first_control_plane.get_host()

        for path, config in cluster.inventory['services']['thirdparties'].items():
            group = cluster.create_group_from_groups_nodes_names(config.get('groups', []), config.get('nodes', []))
            hosts_missing = find_hosts_missing_thirdparty(group, path)
            if hosts_missing:
                broken.append(f"thirdparty {path} is missing on {hosts_missing}")
                # if thirdparty is missing somewhere, do not check anything further for it
                continue

            is_curl = config['source'][:4] == 'http' and '://' in config['source'][4:8]
            expected_sha = None

            # Get sha from source, if it can be downloaded
            if is_curl:
                cluster.log.verbose(f"Thirdparty {path} doesn't have default sha, download it...")
                # Create tmp dir for loading thirdparty without default sha
                random_dir = utils.get_remote_tmp_path()
                final_commands = "rm -r -f %s" % random_dir
                random_path = "%s%s" % (random_dir, path)
                cluster.log.verbose('Temporary path: %s' % random_path)
                remote_commands = "mkdir -p %s" % ('/'.join(random_path.split('/')[:-1]))
                # Load thirdparty to temporary dir
                remote_commands += "&& sudo curl -k -f -g -s --show-error -L %s -o %s" % (config['source'], random_path)
                results = first_control_plane.sudo(remote_commands, pty=True, warn=True)
                if results.is_any_failed():
                    host = first_control_plane_host
                    msg = f"Can`t download thirdparty {path} on {host} for getting sha: {results[host].stdout}"
                    broken.append(msg)
                    cluster.log.verbose(msg)
                else:
                    # Get temporary thirdparty sha
                    cluster.log.verbose(f"Get temporary thirdparty sha for {path}...")
                    results = first_control_plane.sudo(f'openssl sha1 {random_path} | sed "s/^.* //"', warn=True)
                    if results.is_any_failed():
                        host = first_control_plane_host
                        msg = f'failed to get sha for temporary file {random_path} on {host}: {results[host].stderr}'
                        broken.append(msg)
                        cluster.log.verbose(msg)
                    else:
                        expected_sha = results.get_simple_out().strip()
                        cluster.log.verbose(f"Expected sha was got for {path}: {expected_sha}")
                # Remove temporary dir in any case
                cluster.log.verbose(f"Remove temporary dir {random_dir}...")
                first_control_plane.sudo(final_commands, warn=True)
            else:
                script = utils.read_internal(config['source'])
                expected_sha = utils.get_stream_sha1(io.BytesIO(script.encode('utf-8')))

            recommended_sha = thirdparties.get_thirdparty_recommended_sha(path, cluster)
            if recommended_sha is not None and recommended_sha != expected_sha:
                warnings.append(f"{path} source contains not recommended thirdparty version for used kubernetes version")

            if config.get("sha1", expected_sha) != expected_sha and expected_sha is not None:
                broken.append("Given sha is not equal with actual sha from source for %s" % path)

            expected_sha = config.get("sha1", expected_sha)

            if expected_sha is None:
                cluster.log.verbose(f"Can`t get expected sha for {path}, skip it")
                # Skip checking sha if something went wrong or this sha can't be loaded
                continue

            results = group.sudo(f'openssl sha1 {path} | sed "s/^.* //"', warn=True)
            actual_sha: Optional[str] = None
            first_host: Optional[str] = None
            # Searching actual SHA, if possible
            for host, result in results.items():
                if result.failed:
                    broken.append(f'failed to get {path} sha {host}: {result.stderr}')
                    continue

                found_sha = result.stdout.strip()
                if actual_sha is None:
                    actual_sha = found_sha
                    first_host = host
                elif actual_sha != found_sha:
                    broken.append(f'got inconsistent sha for {path}: {found_sha} on host {host}, '
                                  f'different from first host {first_host} sha {actual_sha}')
                    actual_sha = None
                    break

            if actual_sha is None:
                # was not able to find single actual SHA, errors already collected, nothing to do
                continue
            if actual_sha != expected_sha:
                broken.append(f'expected sha {expected_sha} is not equal to actual sha {actual_sha} for {path}')
                continue

            successful.append(path)
            # SHA is correct, now check if it is an archive and if it does, then also check SHA for archive content
            if 'unpack' in config:
                unpack_dir = config['unpack']
                extension = path.split('.')[-1]
                if extension == 'zip': 
                    res = group.sudo(
                                 # for each file in archive
                                 ' unzip -qq -l %s | awk \'NF > 3 { print $4 }\' | while read file_name; do '
                                 '  echo ${file_name} '  # print   1) filename
                                 '    $(sudo unzip -p %s ${file_name} | openssl sha1 | cut -d\\  -f2) '  # 2) sha archive
                                 '    $(sudo openssl sha1 %s/${file_name} | cut -d\\  -f2); '  # 3) sha unpacked
                                 'done' % (path, path, unpack_dir)) 
                else :
                    res = group.sudo('tar tf %s | grep -vw "./" | while read file_name; do '  # for each file in archive
                                 '  echo ${file_name} '  # print   1) filename
                                 '    $(sudo tar xfO %s ${file_name} | openssl sha1 | cut -d\\  -f2) '  # 2) sha archive
                                 '    $(sudo openssl sha1 %s/${file_name} | cut -d\\  -f2); '  # 3) sha unpacked
                                 'done' % (path, path, unpack_dir))
                
                # for each file on each host, verify that SHA in archive is equal to SHA for unpacked
                for host, result in res.items():
                    if result.failed:
                        broken.append(f'can not verify files SHA for archive {path} '
                                      f'on host {host}, unpacked to {unpack_dir}')
                        continue
                    files_results = result.stdout.strip().split('\n')
                    for file_result in files_results:
                        result_parts = file_result.split()
                        if len(result_parts) != 3:
                            broken.append(f'can not verify files SHA for archive {path} '
                                          f'on host {host}, unpacked to {unpack_dir}')
                            continue
                        filename, archive_hash, fs_hash = result_parts[0], result_parts[1], result_parts[2]
                        if archive_hash != fs_hash:
                            broken.append(f'hash for file {filename} from archive {path} '
                                          f'on host {host} is not equal to hash for file unpacked to {unpack_dir}')

        if broken:
            raise TestFailure('Found inconsistent hashes', hint=yaml.safe_dump(broken))
        if warnings:
            raise TestWarn('Found warnings', hint=yaml.safe_dump(warnings))
        tc.success('All found hashes are correct')


def find_hosts_missing_thirdparty(group: NodeGroup, path: str) -> List[str]:
    """
    Search group for a list of hosts where thirdparty is missing
    :param group: group of hosts where to search thirdparty
    :param path: path to thirdparty to search
    :return: list of hosts where thirdparty is missing
    """
    results = group.sudo(f'ls {path}', warn=True)
    missing = []
    for host, result in results.items():
        if result.failed:
            missing.append(host)
    return missing


def kubernetes_nodes_existence(cluster: KubernetesCluster) -> None:
    with TestCase(cluster, '209', "Kubernetes", "Nodes Existence",
                  default_results="All nodes presented"):
        nodes_description = get_nodes_description(cluster)
        nodes_names = kubernetes.get_actual_roles(nodes_description).keys()
        not_found = []
        for node in cluster.inventory['nodes']:
            if 'control-plane' in node['roles'] or 'worker' in node['roles']:
                if node['name'] in nodes_names:
                    cluster.log.debug("Node \"%s\" is found in cluster" % node['name'])
                else:
                    not_found.append(node['name'])
                    cluster.log.error("Node \"%s\" is not found in cluster" % node['name'])
        not_found = list(set(not_found))
        if not_found:
            raise TestFailure("Nodes not found: %s" % ', '.join(not_found),
                              hint="The cluster must contain all the nodes that are described in the inventory. Add "
                                   "the missing nodes to the cluster.")


def kubernetes_nodes_roles(cluster: KubernetesCluster) -> None:
    with TestCase(cluster, '210', "Kubernetes", "Nodes Roles",
                  default_results="All nodes have the correct roles"):
        nodes_description = get_nodes_description(cluster)
        nodes_roles = kubernetes.get_actual_roles(nodes_description)
        nodes_with_bad_roles = []
        for node_name, actual_roles in nodes_roles.items():
            node = cluster.get_node_by_name(node_name)
            if node is None:
                # TODO cluster has unexpected node. Need to add check to kubernetes.nodes.existence task.
                continue

            expected_roles = set(node['roles']).intersection({'control-plane', 'worker'})
            for expected_role in expected_roles:
                if expected_role not in actual_roles:
                    nodes_with_bad_roles.append(node_name)
                    cluster.log.error(f"Node \"{node_name}\" has to be {expected_role}, but has invalid role")
                else:
                    cluster.log.debug(f"Node \"{node_name}\" has correct {expected_role} role")

        nodes_with_bad_roles = list(set(nodes_with_bad_roles))
        if nodes_with_bad_roles:
            raise TestFailure("Incorrect role detected at: %s" % ', '.join(nodes_with_bad_roles),
                              hint="Some nodes whose role differs from that specified in the "
                                   "inventory were detected. The configuration of these nodes should be fixed.")


def kubernetes_nodes_condition(cluster: KubernetesCluster, condition_type: str) -> None:
    with TestCase(cluster, '211', "Kubernetes", "Nodes Condition - %s" % condition_type) as tc:
        nodes_description = get_nodes_description(cluster)
        expected_status = 'False'
        if condition_type == 'Ready':
            expected_status = 'True'
        positive_conditions = []
        negative_conditions = []
        nodes_conditions = kubernetes.get_nodes_conditions(nodes_description)
        for node_name, conditions_by_type in nodes_conditions.items():
            if condition_type in conditions_by_type:
                condition = conditions_by_type[condition_type]
                cluster.log.debug("Node \"%s\" condition \"%s\" is \"%s\""
                                  % (node_name, condition['type'], condition['reason']))
                if condition['status'] != expected_status:
                    negative_conditions.append(condition['reason'])
                else:
                    positive_conditions.append(condition['reason'])
            else:
                raise TestFailure("Failed to detect at %s" % node_name)

        negative_conditions = list(set(negative_conditions))
        positive_conditions = list(set(positive_conditions))

        if negative_conditions:
            raise TestFailure("%s" % ', '.join(negative_conditions),
                              hint="A condition in negative status means that there are problems with the health of "
                                   "the node.")

        tc.success(results="%s" % ', '.join(positive_conditions))


def get_not_running_pods(cluster: KubernetesCluster) -> str:
    # Completed pods should be excluded from the list as well
    get_pods_cmd = ('kubectl get pods -A | grep -v Running '
                    '| awk \'{ print $1" "$2" "$4 }\' | grep -vw Completed || true')
    result = cluster.nodes['control-plane'].get_any_member().sudo(get_pods_cmd)
    cluster.log.verbose(result)
    return list(result.values())[0].stdout.strip()


def kubernetes_pods_condition(cluster: KubernetesCluster) -> None:
    system_namespaces = ["kube-system", "ingress-nginx", "kube-public", "kubernetes-dashboard", "default",
                         "local-path-storage", "calico-apiserver"]
    critical_states = cluster.globals['pods']['critical_states']
    with TestCase(cluster, '207', "Kubernetes", "Pods Condition") as tc:
        pods_descriptions = get_not_running_pods(cluster).split('\n')[1:]
        total_failed_amount = len(pods_descriptions)

        if total_failed_amount > 0:
            cluster.log.debug("Not running pods:\n" + "\n".join(pods_descriptions) + "\n")

        critical_system_failed_amount = 0

        for pod_description in pods_descriptions:
            split_description = pod_description.split(' ')
            if split_description[0] in system_namespaces and split_description[2] in critical_states:
                cluster.log.debug(f"Pod {split_description[1]} is in critical state {split_description[2]}")
                critical_system_failed_amount += 1

        if critical_system_failed_amount > 0:
            s = ''
            if critical_system_failed_amount != 1:
                s = 's'
            raise TestFailure("%s failed system pod%s" % (critical_system_failed_amount, s),
                              hint="Try to determine the cause of the pods failure. Then redeploy, reapply, or restart them. If "
                                   "this is not fixed, the cluster may not work or may work incorrectly.")
        elif total_failed_amount > 0:
            s = ''
            if total_failed_amount != 1:
                s = 's'
            raise TestWarn("%s pod%s are failed/not running" % (total_failed_amount, s),
                           hint="Try to determine the reason the pods are not operational, "
                                "try to wait, redeploy, reapply, or restart them. "
                                "If this is not fixed, some deployed applications may not work or may work incorrectly.")
        else:
            tc.success(results="All pods are running")


def kubernetes_dashboard_status(cluster: KubernetesCluster) -> None:
    with TestCase(cluster, '208', "Plugins", "Dashboard Availability") as tc:
        retries = 10
        test_service_succeeded = False
        test_ingress_succeeded = False
        ingress_ip = cluster.inventory['control_plain']['internal']

        if not cluster.inventory['plugins']['kubernetes-dashboard']['install']:
            tc.success(results="skipped")
        else:
            # check dashboard service 
            cluster.log.debug('Check kubernetes-dashboard service...')
            control_plane = cluster.nodes['control-plane'].get_first_member()
            i = 0
            while not test_service_succeeded and i < retries:
                i += 1
                results = control_plane.sudo(
                    "kubectl get svc -n kubernetes-dashboard kubernetes-dashboard "
                    "-o=jsonpath=\"{['spec.clusterIP']}\"", warn=True)
                for result in results.values():
                    if result.failed:
                        cluster.log.debug(f'Can not get dashboard service IP: {result.stderr} ')
                        raise TestFailure("not available",
                                          hint=f"Please verify the following Kubernetes Dashboard status and fix this issue")
                found_url = results.get_simple_out()
                if ipaddress.ip_address(found_url).version == 4:
                    check_url = control_plane.sudo(
                        f'curl -k -I https://{found_url}:443 -s -S  -w "%{{http_code}}"', pty=True, warn=True)
                else:
                    check_url = control_plane.sudo(
                        f'curl -g -k -I https://[{found_url}]:443 -s -S -w "%{{http_code}}"', pty=True, warn=True)
                status = list(check_url.values())[0].stdout
                lst = status.split('\n')
                if lst[len(lst)-1] == '200':
                    cluster.log.verbose(status)
                    cluster.log.debug('Dashboard service is OK')
                    test_service_succeeded = True
                else:
                    cluster.log.debug(f'Dashboard service is not running yet... Retries left: {retries - i}')
                    time.sleep(60)

            # check dashboard ingress
            cluster.log.debug('Check kubernetes-dashboard ingress...')
            i = 0
            while not test_ingress_succeeded and test_service_succeeded and i < retries:
                i += 1
                results = control_plane.sudo(
                    "kubectl -n kubernetes-dashboard get ingress kubernetes-dashboard "
                    "-o=jsonpath=\"{.spec.rules[0].host}\"", warn=True)
                for result in results.values():
                    if result.failed:
                        cluster.log.debug(f'Can not get dashboard ingress hostname: {result.stderr} ')
                        raise TestFailure("not available",
                                          hint=f"Please verify the following Kubernetes Dashboard status and fix this issue")
                found_url = results.get_simple_out()
                
                if ipaddress.ip_address(ingress_ip).version == 4:
                    check_url = control_plane.sudo(
                        f'curl -k -I -L --resolve {found_url}:443:{ingress_ip} https://{found_url} '
                        f'-s -S -w "%{{http_code}}"', pty=True, warn=True)
                else:
                    check_url = control_plane.sudo(
                        f'curl -g -k -I -L --resolve "{found_url}:443:{ingress_ip}" https://{found_url} '
                        f'-s -S -w "%{{http_code}}"', pty=True, warn=True)
                status = list(check_url.values())[0].stdout
                lst = status.split('\n')
                if lst[len(lst)-1] == '200':
                    cluster.log.verbose(status)
                    cluster.log.debug('Dashboard ingress is OK')
                    test_ingress_succeeded = True
                else:
                    cluster.log.debug(f'Dashboard ingress is not running yet... Retries left: {retries - i}')
                    time.sleep(60)

            # test results
            if test_service_succeeded and test_ingress_succeeded:
                tc.success(results="available")
            else:
                raise TestFailure("not available",
                              hint=f"Please verify the following Kubernetes Dashboard status and fix this issue:\n{status}")


def kubernetes_audit_policy_configuration(cluster: KubernetesCluster) -> None:
    with TestCase(cluster, '229', "Kubernetes", "Audit policy configuration") as tc:
        expected_config = yaml.dump(cluster.inventory['services']['audit']['cluster_policy'])

        api_server_extra_args = cluster.inventory['services']['kubeadm']['apiServer']['extraArgs']
        audit_file_name = api_server_extra_args['audit-policy-file']

        control_planes = cluster.nodes['control-plane']

        broken = []
        result = control_planes.sudo(f"cat {audit_file_name}", warn=True)
        for node in control_planes.get_ordered_members_list():
            policy_result = result[node.get_host()]
            node_name = node.get_node_name()
            if policy_result.failed:
                broken.append(f"{node_name}: {audit_file_name} is absent")
            else:
                actual_config = policy_result.stdout
                diff = utils.get_yaml_diff(
                    actual_config, expected_config,
                    fromfile=audit_file_name,
                    tofile="inventory['services']['audit']['cluster_policy']")
                if diff is not None:
                    cluster.log.debug(f"Configuration of audit policy is not actual on {node_name} node")
                    cluster.log.debug(diff)
                    broken.append(f"{node_name}: {audit_file_name} is not actual")

        if broken:
            broken.append(
                "Check DEBUG logs for details. "
                "To have the check passed, you may need to call `kubemarine install --tasks deploy.kubernetes.audit` "
                "or change the inventory file accordingly."
            )
            raise TestFailure('invalid', hint=yaml.safe_dump(broken))
        else:
            tc.success(results='valid')


def nodes_pid_max(cluster: KubernetesCluster) -> None:
    with TestCase(cluster, '202', "Nodes", "Nodes pid_max correctly installed") as tc:
        control_plane = cluster.nodes['control-plane'].get_any_member()
        yaml = ruamel.yaml.YAML()
        nodes_failed_pid_max_check = {}
        nodes_warned_pid_max_check = {}
        for node in cluster.make_group_from_roles(['control-plane', 'worker']).get_ordered_members_list():
            node_name = node.get_node_name()

            node_info = control_plane.sudo("kubectl get node %s -o yaml" % node_name).get_simple_out()
            config = yaml.load(node_info)
            max_pods = int(config['status']['capacity']['pods'])

            pid_max = int(node.sudo("cat /proc/sys/kernel/pid_max").get_simple_out())

            kubelet_config = node.sudo("cat /var/lib/kubelet/config.yaml").get_simple_out()
            config = yaml.load(kubelet_config)

            if 'podPidsLimit' in config:
                pod_pids_limit = int(config['podPidsLimit'])

                # we need limited podPidsLimit to avoid PIDs exhaustion
                if pod_pids_limit != -1: 
                    required_pid_max = max_pods * pod_pids_limit + 2048
                    cluster.log.debug("Current values:\n maxPods = %s \n podPidsLimit = %s \n pid_max = %s"
                                  % (max_pods, pod_pids_limit, pid_max))
                    cluster.log.debug("Required pid_max for current kubelet configuration is %s for node '%s'"
                                  % (required_pid_max, node_name))
                    inventory_pid_max = cast(int, sysctl.get_parameter(cluster, node, 'kernel.pid_max'))
                    if pid_max != inventory_pid_max:
                        nodes_warned_pid_max_check[node_name] = [pid_max, inventory_pid_max]
                    if pid_max < required_pid_max:
                        nodes_failed_pid_max_check[node_name] = [pid_max, required_pid_max]
                else:
                    cluster.log.error("podPidsLimit is set to unlimited in the /var/lib/kubelet/config.yaml "
                                      f"for node '{node_name}'")
                    nodes_failed_pid_max_check[node_name] = [pid_max, -1]
            else:
                cluster.log.error("No podPidsLimit set in the /var/lib/kubelet/config.yaml for node '%s'" % node_name)
                nodes_failed_pid_max_check[node_name] = [pid_max, -1]

        if nodes_failed_pid_max_check:
            output = "The requirement for the 'pid_max' value is not met for nodes:\n"
            for node_name, (pid_max, required_pid_max) in nodes_failed_pid_max_check.items():
                if required_pid_max == -1:
                    output += ("For node %s podPidsLimit is unlimited\n" % node_name)
                else:
                    output += ("For node %s pid_max value = '%s', but it should be >= then '%s'\n"
                               % (node_name, pid_max, required_pid_max))
            raise TestFailure('invalid', hint=output)

        if nodes_warned_pid_max_check:
            output = ''
            for node_name, (pid_max, inventory_pid_max) in nodes_warned_pid_max_check.items():
                output += ("For node %s the 'kernel.pid_max' value defined in system = %s, "
                           "but 'kernel.pid_max', which defined in cluster.yaml = %s\n"
                           % (node_name, pid_max, inventory_pid_max))

            raise TestWarn('found warnings', hint=output)

        tc.success(results="pid_max correctly installed on all nodes")


def verify_selinux_status(cluster: KubernetesCluster) -> None:
    """
    This method is a test, which checks the status of Selinux. It must be `enforcing`. It may be `permissive`, but must
    be explicitly specified in the inventory. Otherwise, the test will fail. This test is applicable only for systems of
    the RHEL family.
    :param cluster: KubernetesCluster object
    :return: None
    """
    with TestCase(cluster, '213', "Security", "Selinux security policy") as tc:
        group = cluster.nodes['all'].get_subgroup_with_os(['rhel', 'rhel8', 'rhel9'])
        if group.is_empty():
            return tc.success("No RHEL nodes found")
        _, selinux_result, selinux_parsed_result = \
            selinux.is_config_valid(group,
                                    state=selinux.get_expected_state(cluster.inventory),
                                    policy=selinux.get_expected_policy(cluster.inventory),
                                    permissive=selinux.get_expected_permissive(cluster.inventory))
        cluster.log.debug(selinux_result)
        enforcing_ips = []
        permissive_ips = []
        bad_ips = []
        for host, results in selinux_parsed_result.items():
            if results.get('status', '') != 'disabled':
                if results['mode'] == 'enforcing':
                    enforcing_ips.append(host)
                elif results['mode'] == 'permissive' and cluster.inventory.get('services', {})\
                        .get('kernel_security', {}).get('selinux', {}).get('state') == 'permissive':
                    permissive_ips.append(host)
                else:
                    bad_ips.append([host, results['mode']])
            else:
                bad_ips.append([host, 'disabled'])

        if group.nodes_amount() == len(enforcing_ips):
            tc.success(results='enforcing')
        elif len(bad_ips) == 0:
            pretty_list = '\n - ' + ('\n - '.join(permissive_ips))
            raise TestWarn('permissive',
                           hint=f"It is not recommended to use the permissive state, but this is possible if you "
                                f"explicitly specify this in your inventory, thereby assuming all risks. Only "
                                f"\"enforcing\" policy is recommended. Please use it on the following "
                                f"nodes:{pretty_list}")
        else:
            bad_states = []
            pretty_list_ips = []
            for ip_state in bad_ips:
                ip = ip_state[0]
                state = ip_state[1]
                if state not in bad_states:
                    bad_states.append(state)
                if ip not in pretty_list_ips:
                    pretty_list_ips.append(ip)
            pretty_list = '\n - ' + ('\n - '.join(pretty_list_ips))
            raise TestFailure(', '.join(bad_states),
                              hint=f"Selinux is configured with the wrong state, which is not recommended. Only "
                                   f"\"enforcing\" policy is recommended. Please use it on the following "
                                   f"nodes:{pretty_list}")


def verify_selinux_config(cluster: KubernetesCluster) -> None:
    """
    This method is a test, which compares the configuration of Selinux on the nodes with the configuration specified in
    the inventory or with the one by default. If the configuration does not match, the test will fail.
    :param cluster: KubernetesCluster object
    :return: None
    """
    with TestCase(cluster, '214', "Security", "Selinux configuration") as tc:
        group = cluster.nodes['all'].get_subgroup_with_os(['rhel', 'rhel8', 'rhel9'])
        if group.is_empty():
            return tc.success("No RHEL nodes found")
        selinux_configured, selinux_result, _ = \
            selinux.is_config_valid(group,
                                    state=selinux.get_expected_state(cluster.inventory),
                                    policy=selinux.get_expected_policy(cluster.inventory),
                                    permissive=selinux.get_expected_permissive(cluster.inventory))
        cluster.log.debug(selinux_result)
        if selinux_configured:
            tc.success(results='valid')
        else:
            raise TestFailure('invalid',
                              hint=f"Selinux is incorrectly configured - its configuration is different from the one "
                                   f"specified in the inventory. Check the configuration and run the selinux setup task"
                                   f" using the installation procedure.")


def verify_firewalld_status(cluster: KubernetesCluster) -> None:
    """
    This method is a test, which verifies that the FirewallD is disabled on cluster nodes, otherwise the test will fail.
    :param cluster: KubernetesCluster object
    :return: None
    """
    with TestCase(cluster, '215', "Security", "Firewalld status") as tc:
        group = cluster.nodes['all']
        firewalld_disabled, firewalld_result = system.is_firewalld_disabled(group)
        cluster.log.debug(firewalld_result)
        if firewalld_disabled:
            tc.success(results='disabled')
        else:
            raise TestFailure('enabled',
                              hint=f"FirewallD must be disabled as it is not supported and can create compatibility "
                                   f"issues. To solve this problem, execute the firewalld disable task in the installation "
                                   f"procedure.")


def verify_time_sync(cluster: KubernetesCluster) -> None:
    """
    This method is a test that verifies that the time between all nodes does not lag behind.
    :param cluster: KubernetesCluster object
    :return: None
    """
    with TestCase(cluster, '218', "System", "Time difference") as tc:
        group = cluster.nodes['all']
        current_node_time, nodes_timestamp, time_diff = system.get_nodes_time(group)
        cluster.log.verbose('Current node time: %s' % current_node_time)
        cluster.log.verbose('Time difference: %s' % time_diff)
        cluster.log.verbose('Nodes time details:')
        for host, timestamp in nodes_timestamp.items():
            cluster.log.verbose(' - %s: %s' % (host, timestamp))

        if time_diff > cluster.globals['nodes']['max_time_difference']:
            raise TestWarn("%sms" % time_diff,
                           hint=f"The time difference between nodes is too large, this can lead to incorrect "
                                f"behavior of Kubernetes and services. To fix this problem, run the NTP configuring "
                                f"task on all nodes with the correct parameters. It is also worth paying attention to "
                                f"the delay between the deployed node and all the others - too much delay can lead to "
                                f"incorrect time measurements.")
        tc.success(results="%sms" % time_diff)


def verify_swap_state(cluster: KubernetesCluster) -> None:
    """
    This method is a test, which verifies that swap is disabled on all nodes in the cluster, otherwise the test will
    fail.
    :param cluster: KubernetesCluster object
    :return: None
    """
    with TestCase(cluster, '216', "System", "Swap state") as tc:
        group = cluster.nodes['all']
        swap_disabled, swap_result = system.is_swap_disabled(group)
        cluster.log.debug(swap_result)
        if swap_disabled:
            tc.success(results='disabled')
        else:
            raise TestFailure('enabled',
                              hint=f"Swap must be disabled as it is not supported and can create performance "
                                   f"issues. To solve this problem, execute the swap disable task in the installation "
                                   f"procedure.")


def verify_modprobe_rules(cluster: KubernetesCluster) -> None:
    """
    This method is a test, which compares the modprobe rules on the nodes with the rules specified in the inventory or
    with default rules. If rules does not match, the test will fail.
    :param cluster: KubernetesCluster object
    :return: None
    """
    with TestCase(cluster, '217', "System", "Modprobe rules") as tc:
        group = cluster.nodes['all']
        modprobe_valid, _, modprobe_result = modprobe.is_modprobe_valid(group)
        cluster.log.debug(modprobe_result)
        if modprobe_valid:
            tc.success(results='valid')
        else:
            raise TestFailure('invalid',
                              hint=f"Modprobe rules do not match those loaded in modprobe on cluster nodes. Check "
                                   f"manually what the differences are and make changes on the appropriate nodes.")


def verify_sysctl_config(cluster: KubernetesCluster) -> None:
    """
    This test compares the kernel parameters on the nodes
    with the parameters specified in the inventory or with the default parameters.
    If the configured parameters are not presented, the test fails.

    :param cluster: KubernetesCluster object
    :return: None
    """
    with TestCase(cluster, '232', "System", "Kernel Parameters") as tc:
        group = cluster.nodes['all']
        sysctl_valid = sysctl.is_valid(group)
        if sysctl_valid:
            cluster.log.debug("Required kernel parameters are presented")
            tc.success(results='valid')
        else:
            raise TestFailure('invalid',
                              hint=f"Some configured kernel parameters are not loaded on the cluster nodes.\n"
                                   f"Check manually what the differences are, and make changes on the appropriate nodes.")


def verify_system_audit_rules(cluster: KubernetesCluster) -> None:
    """
    This test compares the Audit rules on the nodes
    with the rules specified in the inventory or with the default rules.
    If the configured rules are not presented, the test fails.

    :param cluster: KubernetesCluster object
    :return: None
    """
    with TestCase(cluster, '231', "System", "Audit Daemon Rules") as tc:
        group = cluster.make_group_from_roles(['control-plane', 'worker'])
        rules_valid, auditctl_results = audit.audit_rules_valid(group)
        cluster.log.debug(auditctl_results)
        if rules_valid:
            tc.success(results='valid')
        else:
            raise TestFailure('invalid',
                              hint=f"Some configured Audit rules are not loaded on the cluster nodes.\n"
                                   f"Check manually what the differences are, and make changes on the appropriate nodes.")


def etcd_health_status(cluster: KubernetesCluster) -> None:
    """
    This method is a test, check ETCD health
    :param cluster: KubernetesCluster object
    :return: None
    """
    with TestCase(cluster, '219', "ETCD", "Health status ETCD") as tc:
        try:
            etcd.wait_for_health(cluster, cluster.nodes['control-plane'].get_any_member())
        except Exception as e:
            cluster.log.verbose('Failed to load and parse ETCD status')
            raise TestFailure('invalid',
                              hint=f"ETCD not ready, please check"
                                   f" because of {e} ") from None
        tc.success(results='healthy')


def container_runtime_configuration_check(cluster: KubernetesCluster) -> None:
    with TestCase(cluster, '204', "Services", f"Containerd Configuration Check") as tc:
        expected_config = containerd.get_config_as_toml(cluster.inventory['services']['cri']['containerdConfig'])
        expected_registries = {registry: containerd.get_config_as_toml(registry_conf)
                               for registry, registry_conf in
                               cluster.inventory['services']['cri'].get('containerdRegistriesConfig', {}).items()}
        kubernetes_nodes = cluster.make_group_from_roles(['control-plane', 'worker'])
        actual_configs, actual_registries = containerd.fetch_containerd_config(kubernetes_nodes)

        success = True
        for node in kubernetes_nodes.get_ordered_members_list():
            actual_config = actual_configs[node.get_host()]
            diff = DeepDiff(actual_config, expected_config)
            if diff:
                cluster.log.debug(f"Configuration of containerd is not actual on {node.get_node_name()} node")
                utils.print_diff(cluster.log, diff)
                success = False

            diff = DeepDiff(actual_registries.get(node.get_host(), {}), expected_registries)
            if diff:
                cluster.log.debug(f"Configuration of containerd registries is not actual on {node.get_node_name()} node")
                utils.print_diff(cluster.log, diff)
                success = False

        if not success:
            message = dedent(
                """\
                The configuration of the container runtime does not match
                the effectively resolved configuration from the inventory.
                Check DEBUG logs for details.
                To have the check passed, you may need to call
                `kubemarine install --tasks prepare.cri.configure`
                or change the inventory file accordingly.
                """.rstrip())
            raise TestFailure('invalid', hint=message)
        else:
            tc.success(results='valid')


def control_plane_configuration_status(cluster: KubernetesCluster) -> None:
    '''
    This test verifies the consistency of the configuration of static pods of Control Plain
    for `kube-apiserver`, `kube-controller-manager`, `kube-scheduler`, and `etcd`.
    :param cluster: KubernetesCluster object
    :return: None
    '''
    with TestCase(cluster, '220', "Control plane", "configuration status") as tc:
        messages = []

        cluster.log.debug("Checking consistency with kubeadm-config ConfigMap")
        failed_nodes = _control_plane_compare_manifests(cluster, with_inventory=False)
        if failed_nodes:
            messages.append(f"Static pod manifests are not consistent with kubeadm-config ConfigMap "
                            f"on control-planes {', '.join(failed_nodes)}")

        cluster.log.debug("Checking consistency with inventory")
        failed_nodes = _control_plane_compare_manifests(cluster, with_inventory=True)
        if failed_nodes:
            messages.append(f"Static pod manifests are not consistent with inventory "
                            f"on control-planes {', '.join(failed_nodes)}")

        if not messages:
            tc.success(results='valid')
        else:
            messages.append(
                "Check DEBUG logs for details. To have the check passed, "
                "you may need to call `kubemarine reconfigure --tasks deploy.kubernetes.reconfigure`, "
                "or change the inventory file accordingly."
            )
            raise TestFailure('invalid', hint=yaml.safe_dump(messages))


def _control_plane_compare_manifests(cluster: KubernetesCluster, *, with_inventory: bool) -> List[str]:
    manifest_diffs = components.compare_manifests(cluster, with_inventory=with_inventory)
    failed_nodes = OrderedSet[str]()
    failed_manifests = set()
    for host, manifests in manifest_diffs.items():
        for manifest_, diff in manifests.items():
            if diff is None:
                continue

            node_name = cluster.get_node_name(host)
            cluster.log.debug(f"Static pod manifest /etc/kubernetes/manifests/{manifest_}.yaml is not actual "
                              f"on control-plane {node_name}")
            if manifest_ in failed_manifests:
                cluster.log.debug('<truncated, enable verbose logs for details>')
                cluster.log.verbose(diff + '\n')
            else:
                failed_manifests.add(manifest_)
                cluster.log.debug(diff + '\n')

            failed_nodes.add(node_name)

    return list(failed_nodes)


def control_plane_health_status(cluster: KubernetesCluster) -> None:
    '''
    This test verifies the health of static pods `kube-apiserver`, `kube-controller-manager`,
    `kube-scheduler`, and `etcd`.
    :param cluster: KubernetesCluster object
    :return: None
    '''
    with TestCase(cluster, '221', "Control plane", "health status") as tc:
        static_pods = ['kube-apiserver', 'kube-controller-manager', 'kube-scheduler', 'etcd']
        static_pod_names = []

        for control_plane in cluster.nodes['control-plane'].get_ordered_members_list():
            for static_pod in static_pods:
                static_pod_names.append(static_pod + '-' + control_plane.get_node_name())

        first_control_plane = cluster.nodes['control-plane'].get_first_member()
        not_found_pod = []
        for static_pod_name in static_pod_names:
            results = first_control_plane.sudo(f"kubectl get pod -n kube-system -oyaml {static_pod_name}", warn=True)
            exit_code = list(results.values())[0].exited
            if exit_code == 0:
                result = yaml.safe_load(results.get_simple_out())
                if result['status']['containerStatuses'][0]['state'].get('running'):
                    break
            not_found_pod.append(static_pod_name)

        if len(not_found_pod) == 0:
            tc.success(results='valid')
        else:
            raise TestFailure('invalid', hint=f"{not_found_pod} pods doesn't running")


def default_services_configuration_status(cluster: KubernetesCluster) -> None:
    '''
    In this test, the versions of the images of the default services
    `kube-proxy`, `coredns`,
    `calico-node`, `calico-kube-controllers`, `calico-typha`, `calico-apiserver`,
    `ingress-nginx-controller`, `local-path-provisioner`,
    `kubernetes-dashboard`, and `dashboard-metrics-scraper` are checked,
    and the `coredns` configmap is also checked.

    :param cluster: KubernetesCluster object
    :return: None
    '''
    with TestCase(cluster, '222', "Default services", "configuration status") as tc:
        first_control_plane = cluster.nodes['control-plane'].get_first_member()
        original_coredns_cm = yaml.safe_load(generate_configmap(cluster.inventory))
        result = first_control_plane.sudo('kubectl get cm coredns -n kube-system -oyaml')
        coredns_cm = yaml.safe_load(result.get_simple_out())
        diff = utils.get_unified_diff(
            coredns_cm['data']['Corefile'], original_coredns_cm['data']['Corefile'],
            fromfile='kube-system/configmaps/coredns/data/Corefile',
            tofile="inventory['services']['coredns']['configmap']['Corefile']")

        failed_messages = []
        warn_messages = []
        if diff is not None:
            failed_messages.append(f"CoreDNS config is outdated: \n{diff}")

        software_compatibility = static.GLOBALS["compatibility_map"]["software"]
        version = cluster.inventory['services']['kubeadm']['kubernetesVersion']
        coredns_version = software_compatibility["coredns/coredns"][version]["version"]

        entities_to_check = {"kube-system": {"DaemonSet": {"kube-proxy": {"version": version}},
                                             "Deployment": {"coredns": {"version": coredns_version}}}}

        for plugin in plugins.oob_plugins:
            plugin_item = cluster.inventory['plugins'][plugin]
            if not plugin_item['install']:
                continue

            recommended_version = software_compatibility[plugin][version]["version"]
            expected_version = plugin_item['version']
            if recommended_version != expected_version:
                warn_messages.append(f"{plugin!r} has not recommended {expected_version} version. "
                                     f"Recommended: {recommended_version}")

            if plugin == 'calico':
                entities_to_check["kube-system"]["DaemonSet"]["calico-node"] = {"version": expected_version}
                entities_to_check["kube-system"]["Deployment"]["calico-kube-controllers"] = {"version": expected_version}
                if calico.is_apiserver_enabled(cluster.inventory):
                    entities_to_check['calico-apiserver'] = {"Deployment": {"calico-apiserver": {"version": expected_version}}}
                if calico.is_typha_enabled(cluster.inventory):
                    entities_to_check["kube-system"]["Deployment"]["calico-typha"] = {"version": expected_version}

            if plugin == 'nginx-ingress-controller':
                entities_to_check['ingress-nginx'] = {"DaemonSet": {"ingress-nginx-controller": {"version": expected_version}}}
            if plugin == 'local-path-provisioner':
                entities_to_check['local-path-storage'] = {"Deployment": {
                    "local-path-provisioner": {"version": expected_version}}}
            if plugin == 'kubernetes-dashboard':
                image = plugin_item['metrics-scraper']['image']
                image = image.split('@sha256:')[0]
                expected_metrics_scrapper_version = image.split(':')[-1]
                entities_to_check['kubernetes-dashboard'] = {"Deployment": {
                    "kubernetes-dashboard": {"version": expected_version},
                    "dashboard-metrics-scraper": {"version": expected_metrics_scrapper_version}}}

        for namespace, types_dict in entities_to_check.items():
            for kind, services in types_dict.items():
                for service_name, properties in services.items():
                    k8s_object = KubernetesObject(cluster, kind, service_name, namespace)
                    k8s_object.reload(control_plane=first_control_plane, suppress_exceptions=True)
                    if not k8s_object.is_reloaded():
                        stderr = str(k8s_object).rstrip('\n')
                        failed_messages.append(f"failed to load {service_name}: {stderr}")
                        continue

                    if properties["version"] not in k8s_object.obj["spec"]["template"]["spec"]["containers"][0].get("image", ""):
                        failed_messages.append(f"{service_name} has outdated image version")

        if failed_messages:
            raise TestFailure('invalid', hint="\n".join(failed_messages))
        elif warn_messages:
            raise TestWarn('found warnings', hint="\n".join(warn_messages))
        else:
            tc.success(results='valid')


def default_services_health_status(cluster: KubernetesCluster) -> None:
    '''
    This test verifies the health of
    `kube-proxy`, `coredns`,
    `calico-node`, `calico-kube-controllers`, `calico-typha`, `calico-apiserver`,
    `ingress-nginx-controller`, `local-path-provisioner`,
    `kubernetes-dashboard`, and `dashboard-metrics-scraper` pods.

    :param cluster: KubernetesCluster object
    :return: None
    '''
    with TestCase(cluster, '223', "Default services", "health status") as tc:
        entities_to_check = {"kube-system": {"DaemonSet": ["kube-proxy"],
                                             "Deployment": ["coredns"]}}

        for plugin in plugins.oob_plugins:
            plugin_item = cluster.inventory['plugins'][plugin]
            if not plugin_item['install']:
                break

            if plugin == 'calico':
                entities_to_check["kube-system"]["DaemonSet"].append("calico-node")
                entities_to_check["kube-system"]["Deployment"].append("calico-kube-controllers")
                if calico.is_apiserver_enabled(cluster.inventory):
                    entities_to_check['calico-apiserver'] = {"Deployment": ["calico-apiserver"]}
                if calico.is_typha_enabled(cluster.inventory):
                    entities_to_check["kube-system"]["Deployment"].append("calico-typha")

            if plugin == 'ingress-nginx-controller':
                entities_to_check['ingress-nginx'] = {"DaemonSet": ["ingress-nginx-controller"]}
            if plugin == 'local-path-provisioner':
                entities_to_check['local-path-storage'] = {"Deployment": ["local-path-provisioner"]}
            if plugin == 'kubernetes-dashboard':
                entities_to_check['kubernetes-dashboard'] = {"Deployment": [
                    "kubernetes-dashboard", "dashboard-metrics-scraper"]}

        first_control_plane = cluster.nodes['control-plane'].get_first_member()
        not_ready_entities = []
        for namespace, types_dict in entities_to_check.items():
            for kind, services in types_dict.items():
                if kind == 'DaemonSet':
                    for service in services:
                        daemon_set = DaemonSet(cluster, name=service, namespace=namespace)
                        daemon_set.reload(control_plane=first_control_plane, suppress_exceptions=True)
                        if not daemon_set.is_actual_and_ready():
                            not_ready_entities.append(service)
                elif kind == 'Deployment':
                    for service in services:
                        deployment = Deployment(cluster, name=service, namespace=namespace)
                        deployment.reload(control_plane=first_control_plane, suppress_exceptions=True)
                        if not daemon_set.is_actual_and_ready():
                            not_ready_entities.append(service)
        if len(not_ready_entities) == 0:
            tc.success(results='valid')
        else:
            raise TestFailure('invalid', hint=f"{not_ready_entities} pods doesn't ready")


def _check_calico_env(cluster: KubernetesCluster, manifest_: manifest.Manifest) -> List[str]:
    first_control_plane = cluster.nodes['control-plane'].get_first_member()
    failed_messages = []

    def get_envs(obj: dict) -> Dict[str, Union[str, dict]]:
        raw_envs = obj["spec"]["template"]["spec"]["containers"][0]["env"]
        envs = {}
        for env in raw_envs:
            envs[env['name']] = env.get('valueFrom', env.get('value', ''))

        return envs

    entities_to_check = [('DaemonSet', 'calico-node')]
    if calico.is_typha_enabled(cluster.inventory):
        entities_to_check.append(('Deployment', 'calico-typha'))

    for type_, service_name in entities_to_check:
        k8s_object = KubernetesObject(cluster, type_, service_name, 'kube-system')
        k8s_object.reload(control_plane=first_control_plane, suppress_exceptions=True)
        if not k8s_object.is_reloaded():
            stderr = str(k8s_object).rstrip('\n')
            failed_messages.append(f"failed to load {service_name}: {stderr}")
            continue

        actual_env = get_envs(k8s_object.obj)

        expected_obj = manifest_.get_obj(f"{type_}_{service_name}", patch=False)
        buf = io.StringIO()
        utils.yaml_structure_preserver().dump(expected_obj, buf)
        expected_obj = yaml.safe_load(buf.getvalue())
        expected_env = get_envs(expected_obj)

        diff = DeepDiff(actual_env, expected_env)
        if diff:
            cluster.log.debug(f"{service_name!r} env configuration is outdated")
            utils.print_diff(cluster.log, diff)
            failed_messages.append(f"{service_name!r} env configuration is outdated. Check DEBUG logs for details.")

    return failed_messages


def calico_config_check(cluster: KubernetesCluster) -> None:
    '''
    This test checks the configuration of the `calico-node` envs,
    Calico's ConfigMap in case of `ipam`, and also performed `calicoctl ipam check`.
    :param cluster: KubernetesCluster object
    :return: None
    '''
    with TestCase(cluster, '224', "Calico", "configuration check") as tc:
        first_control_plane = cluster.nodes['control-plane'].get_first_member()
        failed_messages = []
        warn_messages = []

        # Check calico-node and calico-typha env variables.
        # Since they are relatively complex to calculate, let's use real enriched manifest.
        processor = builtin.get_manifest_processor(cluster, manifest.Identity('calico'))
        if processor is None:
            warn_messages.append("Calico manifest is not installed using default procedure. "
                                 "Cannot check configuration of env variables.")
        else:
            failed_messages.extend(_check_calico_env(cluster, processor.enrich()))

        # Check calico ipam config of CNI config
        result = first_control_plane.sudo(f"kubectl get cm calico-config -n kube-system -oyaml")
        calico_config = yaml.safe_load(result.get_simple_out())
        cni_network_config = yaml.safe_load(calico_config["data"]["cni_network_config"])
        ip = cluster.inventory['services']['kubeadm']['networking']['podSubnet'].split('/')[0]
        if utils.isipv(ip, [4]):
            ipam_config = cluster.inventory["plugins"]["calico"]["cni"]["ipam"]["ipv4"]
        else:
            ipam_config = cluster.inventory["plugins"]["calico"]["cni"]["ipam"]["ipv6"]
        ddiff = DeepDiff(ipam_config, cni_network_config["plugins"][0]["ipam"], ignore_order=True)
        if ddiff:
            cluster.log.debug("'cni_network_config.plugins[0].ipam' of calico-config ConfigMap is outdated:")
            utils.print_diff(cluster.log, ddiff)
            failed_messages.append("'cni_network_config.plugins[0].ipam' of calico-config ConfigMap is outdated. "
                                   "Check DEBUG logs for details.")

        # Check calicoctl and calico version match, and check ipam problems.
        collector = CollectorCallback(cluster)
        first_control_plane.sudo(
            "calicoctl ipam check", pty=True, warn=True, callback=collector)

        ipam_check = collector.result.get_simple_out()
        problems = re.findall(r'found (\d+) problems', ipam_check)
        if 'Version mismatch' in ipam_check:
            version_mismatch_lines = ipam_check.split('\n')

            def calico_version(criteria: str) -> str:
                return next(map(lambda l: l.split()[-1],
                                filter(lambda l: criteria in l,
                                       version_mismatch_lines),
                                ),
                            'unknown')

            client_version = calico_version('Client Version')
            cluster_version = calico_version('Cluster Version')
            failed_messages.append(f"client version {client_version} mismatches the cluster version {cluster_version}")
        elif problems and int(problems[0]) > 0:
            failed_messages.append("ipam check indicates some problems, "
                                   "for more info you can use `calicoctl ipam check --show-problem-ips`")
        elif collector.result.is_any_failed() or not problems:
            raise GroupResultException(collector.result)
        if failed_messages:
            raise TestFailure('invalid', hint="\n".join(failed_messages))
        elif warn_messages:
            raise TestWarn('found warnings', hint="\n".join(warn_messages))
        else:
            tc.success(results='valid')


def calico_apiserver_health_status(cluster: KubernetesCluster) -> None:
    """
    This test verifies the Calico API server health.

    :param cluster: KubernetesCluster object
    :return: None
    """
    with TestCase(cluster, '230', "Calico", "API server health status") as tc:
        if not calico.is_apiserver_enabled(cluster.inventory):
            return tc.success(results='disabled')

        control_planes = cluster.nodes["control-plane"]
        with kubernetes.local_admin_config(control_planes) as kubeconfig:
            result = control_planes.sudo(
                f"kubectl --kubeconfig {kubeconfig} get ippools.projectcalico.org > /dev/null",
                pty=True, warn=True)

        if result.is_any_failed():
            message = "Calico API server is not ready:\n"
            for host in result.get_failed_hosts_list():
                stderr = result[host].stdout
                if stderr:
                    message += f"{host}: {stderr}"
            raise TestFailure('invalid', hint=message)
        else:
            tc.success(results='valid')


def kubernetes_admission_status(cluster: KubernetesCluster) -> None:
    """
    The method checks status of Pod Security Admissions, default Pod Security Profile,
    and 'kube-apiserver.yaml' and 'kubeadm-config' consistancy
    """
    with TestCase(cluster, '225', "Kubernetes", "Pod Security Admissions") as tc:
        # Consistency of kube-apiserver flags is checked in control_plane.configuration_status
        # Here we check only the admission configuration and global enabled / disabled status.

        control_planes = cluster.nodes['control-plane']
        first_control_plane = control_planes.get_first_member()

        expected_state = cluster.inventory["rbac"]["pss"]["pod-security"]

        kubeadm_config = components.KubeadmConfig(cluster)
        cluster_config = kubeadm_config.load('kubeadm-config', first_control_plane)
        apiserver_actual_args = cluster_config["apiServer"]["extraArgs"]

        actual_state = "disabled"
        if "admission-control-config-file" in apiserver_actual_args and (
                "PodSecurity=true" in apiserver_actual_args.get("feature-gates", "")
                or admission.is_pod_security_unconditional(cluster)
        ):
            actual_state = "enabled"

        if expected_state != actual_state:
            raise TestFailure('invalid',
                              hint=f"Incorrect PSS state. Expected: {expected_state}, actual: {actual_state}")

        if expected_state == "disabled":
            kube_admission_status = 'PSS is "disabled"'
            cluster.log.debug(kube_admission_status)
            return tc.success(results='disabled')

        expected_config = admission.generate_pss(cluster)

        apiserver_expected_args = cluster.inventory['services']['kubeadm']['apiServer']['extraArgs']
        config_file_name = apiserver_expected_args['admission-control-config-file']

        broken = []
        result = control_planes.sudo(f"cat {config_file_name}", warn=True)
        for node in control_planes.get_ordered_members_list():
            config_result = result[node.get_host()]
            node_name = node.get_node_name()
            if config_result.failed:
                broken.append(f"{node_name}: {config_file_name} is absent")
            else:
                actual_config = config_result.stdout
                diff = utils.get_yaml_diff(
                    actual_config, expected_config,
                    fromfile=config_file_name,
                    tofile="inventory['rbac']['pss']")
                if diff is not None:
                    cluster.log.debug(f"Admission configuration is not actual on {node_name} node")
                    cluster.log.debug(diff)
                    broken.append(f"{node_name}: {config_file_name} is not actual")

        if broken:
            broken.append(
                "Check DEBUG logs for details. "
                "To have the check passed, you may need to run `kubemarine manage_pss` with enabled pod-security, "
                "or change the inventory file accordingly."
            )
            raise TestFailure('invalid', hint=yaml.safe_dump(broken))

        kube_admission_status = ('PSS is "enabled", default profile is "%s"'
                                 % cluster.inventory["rbac"]["pss"]["defaults"]["enforce"])
        cluster.log.debug(kube_admission_status)
        tc.success(results='enabled')


def geo_check(cluster: KubernetesCluster) -> None:
    """
    This test checks connectivity between clusters in geo schemas using paas-geo-monitor service.
    This test only work if "procedure.yaml" has "geo-monitor" section filled.
    """
    if not cluster.procedure_inventory or not cluster.procedure_inventory.get("geo-monitor"):
        cluster.log.debug("Geo connectivity check is skipped, no configuration provided")
        return

    status_collected = False
    dns_status: Dict[str, List[str]] = {"failed": []}
    svc_status: Dict[str, List[str]] = {"failed": [], "skipped": []}
    pod_status: Dict[str, List[str]] = {"failed": [], "skipped": []}

    # Here we actually collect information about all statuses, but report information about DNS only.
    # Other statuses are reported in other TestCases below. This is done for better UX.
    with TestCase(cluster, '226', "Geo Monitor", "Geo check - DNS resolving") as tc_dns:
        geo_monitor_inventory = cluster.procedure_inventory["geo-monitor"]
        namespace = geo_monitor_inventory["namespace"]
        service = geo_monitor_inventory["service"]
        control_plane_node = cluster.nodes['control-plane'].get_first_member()

        svc_result = control_plane_node.sudo("kubectl get svc -n %s %s -o yaml" % (namespace, service)).get_simple_out()
        svc = yaml.safe_load(io.StringIO(svc_result))
        ip = svc["spec"]["clusterIP"]
        port = svc["spec"]["ports"][0]["port"]

        # todo: support https?
        status_cmd = f'curl http://{ip}:{port}/peers/status'
        if ipaddress.ip_address(ip).version == 6:
            status_cmd += " -g"
        peers_result = cluster.nodes['control-plane'].get_first_member().\
            sudo(f'curl http://{ip}:{port}/peers/status').get_simple_out()

        peers = yaml.safe_load(io.StringIO(peers_result))
        if len(peers) == 0:
            raise TestFailure("configuration error", hint="geo-monitor instance has no peers")

        for peer in peers:
            status = peer["clusterIpStatus"]
            if not status["dnsStatus"]["resolved"]:
                error = f'FAILED DNS resolving for peer ({peer["name"]}) service ' \
                        f'name: {status["name"]}, error: {status["dnsStatus"]["error"]}'
                dns_status["failed"].append(error)
                svc_status["skipped"].append(error)
                pod_status["skipped"].append(error)
                continue
            if not status["svcStatus"]["available"]:
                error = f'FAILED ping service for peer ({peer["name"]}), ' \
                        f'address: {status["svcStatus"]["address"]}, error: {status["svcStatus"]["error"]}'
                svc_status["failed"].append(error)
                pod_status["skipped"].append(error)
                continue
            if not status["podStatus"]["available"]:
                error = f'FAILED ping pod for peer ({peer["name"]}), ' \
                        f'address: {status["podStatus"]["address"]}, error: {status["podStatus"]["error"]}'
                pod_status["failed"].append(error)
                continue

        status_collected = True
        if dns_status["failed"]:
            raise TestFailure("found failed DNS statuses", hint=yaml.safe_dump(dns_status["failed"]))
        tc_dns.success("all peer names resolved")

    with TestCase(cluster, '226', "Geo Monitor", "Geo check - Pod-to-service") as tc_svc:
        if not status_collected:
            raise TestFailure("configuration error", hint="DNS check failed with error, statuses not collected")

        if svc_status["failed"]:
            raise TestFailure("found unavailable peer services",
                              hint=yaml.safe_dump(svc_status["failed"] + svc_status["skipped"]))
        if svc_status["skipped"]:
            raise TestWarn("found skipped peer services", hint=yaml.safe_dump(svc_status["skipped"]))
        tc_svc.success("all peer services available")

    with TestCase(cluster, '226', "Geo Monitor", "Geo check - Pod-to-pod") as tc_pod:
        if not status_collected:
            raise TestFailure("configuration error", hint="DNS check failed with error, statuses not collected")

        if pod_status["failed"]:
            raise TestFailure("found unavailable peer pod",
                              hint=yaml.safe_dump(pod_status["failed"] + pod_status["skipped"]))
        if pod_status["skipped"]:
            raise TestWarn("found skipped peer pods", hint=yaml.safe_dump(pod_status["skipped"]))
        tc_pod.success("all peer pods available")


def verify_apparmor_status(cluster: KubernetesCluster) -> None:
    """
    This method is a test, which checks the status of Apparmor.
    This test is applicable only for systems of the Debian family.
    :param cluster: KubernetesCluster object
    :return: None
    """
    with TestCase(cluster, '227', "Security", "Apparmor security policy") as tc:
        group = cluster.nodes['all'].get_subgroup_with_os('debian')
        if group.is_empty():
            return tc.success("No Debian nodes found")
        results = group.sudo("aa-enabled")
        enabled_nodes: List[str] = []
        invalid_nodes: List[str] = []
        for host, item in results.items():
            apparmor_status = item.stdout
            cluster.log.warning(f"Apparmor on node: {host} enabled: {apparmor_status}")
            if apparmor_status == "Yes":
                enabled_nodes.append(host)
            else:
                enabled_nodes.append(host)
        if group.nodes_amount() == len(enabled_nodes):
            tc.success(results='enabled')
        else:
            raise TestFailure(f"Apparmor does not properly configured on the following nodes: {invalid_nodes}")


def verify_apparmor_config(cluster: KubernetesCluster) -> None:
    """
    This method tests if the Apparmor configuration matches to 'cluster.yaml' spec.
    This test is applicable only for systems of the Debian family.
    :param cluster: KubernetesCluster object
    :return: None
    """
    with TestCase(cluster, '228', "Security", "Apparmor Configuration") as tc:
        group = cluster.nodes['all'].get_subgroup_with_os('debian')
        if group.is_empty():
            return tc.success("No Debian nodes found")
        expected_profiles = cluster.inventory['services']['kernel_security'].get('apparmor', {})
        if expected_profiles:
            apparmor_configured = apparmor.is_state_valid(group, expected_profiles)
            if apparmor_configured:
                cluster.log.verbose(f"Apparmor is configured properly on cluster")
                tc.success(results='valid')
            else:
                raise TestFailure('invalid',
                        hint=f"Some nodes do not have properly configured Apparmor service")
        else:
            tc.success(results='skipped')


tasks = OrderedDict({
    'services': {
        'security': {
            'selinux': {
                'status': verify_selinux_status,
                'config': verify_selinux_config
            },
            'apparmor': {
                'status': verify_apparmor_status,
                'config': verify_apparmor_config
            },
            'firewalld': {
                'status': verify_firewalld_status
            }
        },
        'system': {
            'time': verify_time_sync,
            'swap': {
                'status': verify_swap_state
            },
            'modprobe': {
                'rules': verify_modprobe_rules
            },
            'sysctl': {
                'config': verify_sysctl_config
            },
            'audit': {
                'rules': verify_system_audit_rules
            },
        },
        'haproxy': {
            'status': lambda cluster: services_status(cluster, 'haproxy'),
        },
        'keepalived': {
            'status': lambda cluster: services_status(cluster, 'keepalived'),
        },
        'container_runtime': {
            'status': lambda cluster: services_status(cluster, 'containerd'),
            'configuration': container_runtime_configuration_check,
        },
        'kubelet': {
            'status': lambda cluster: services_status(cluster, 'kubelet'),
            'pid_max': nodes_pid_max,
            'version': kubelet_version,
            'configuration': kubelet_config,
        },
        'kube-proxy': {
            'configuration': kube_proxy_configmap,
        },
        'packages': {
            'system': {
                'cri_version': lambda cluster: system_packages_versions(cluster, 'containerd'),
                'haproxy_version': lambda cluster: system_packages_versions(cluster, 'haproxy'),
                'keepalived_version': lambda cluster: system_packages_versions(cluster, 'keepalived'),
                'audit_version': lambda cluster: system_packages_versions(cluster, 'audit'),
                'mandatory_versions': mandatory_packages_versions
            },
            'generic': {
                'version': generic_packages_versions
            }
        },
    },
    'thirdparties': {
        'hashes': thirdparties_hashes,
    },
    'kubernetes': {
        'pods': kubernetes_pods_condition,
        'plugins': {
            'dashboard': kubernetes_dashboard_status
        },
        'nodes': {
            'existence': kubernetes_nodes_existence,
            'roles': kubernetes_nodes_roles,
            'condition': {
                "network": lambda cluster: kubernetes_nodes_condition(cluster, 'NetworkUnavailable'),
                "memory": lambda cluster: kubernetes_nodes_condition(cluster, 'MemoryPressure'),
                "disk": lambda cluster: kubernetes_nodes_condition(cluster, 'DiskPressure'),
                "pid": lambda cluster: kubernetes_nodes_condition(cluster, 'PIDPressure'),
                "ready": lambda cluster: kubernetes_nodes_condition(cluster, 'Ready')
            },
        },
        'audit': {
            'policy': kubernetes_audit_policy_configuration,
        },
        'admission': kubernetes_admission_status,
    },
    'etcd': {
        "health_status": etcd_health_status
    },
    'control_plane': {
        "configuration_status": control_plane_configuration_status,
        "health_status": control_plane_health_status
    },
    'default_services': {
        "configuration_status": default_services_configuration_status,
        "health_status": default_services_health_status
    },
    'calico': {
        "config_check": calico_config_check,
        "apiserver": {
            "health_status": calico_apiserver_health_status,
        }
    },
    'geo_check': geo_check,
})


class PaasAction(flow.TasksAction):
    def __init__(self) -> None:
        super().__init__('check paas', tasks)


def create_context(cli_arguments: List[str] = None) -> dict:
    cli_help = '''
    Script for checking Kubernetes cluster PAAS layer.
    
    Hot to use:

    '''

    parser = flow.new_procedure_parser(cli_help, optional_config=True, tasks=tasks)

    parser.add_argument('--csv-report',
                        default='report.csv',
                        help='define CSV report file location')

    parser.add_argument('--csv-report-delimiter',
                        default=';',
                        help='define delimiter type for CSV report')

    parser.add_argument('--html-report',
                        default='report.html',
                        help='define HTML report file location')

    parser.add_argument('--disable-csv-report',
                        action='store_true',
                        help='forcibly disable CSV report file creation')

    parser.add_argument('--disable-html-report',
                        action='store_true',
                        help='forcibly disable HTML report file creation')

    context = flow.create_context(parser, cli_arguments, procedure='check_paas')
    context['testsuite'] = TestSuite()
    context['preserve_inventory'] = False
    context['result'].append('testsuite')

    return context


def main(cli_arguments: List[str] = None) -> TestSuite:
    context = create_context(cli_arguments)
    flow_ = flow.ActionsFlow([PaasAction()])
    result = flow_.run_flow(context, print_summary=False)

    testsuite: TestSuite = result.context['testsuite']

    # Final summary should be printed only to stdout with custom formatting
    # If tests results required for parsing, they can be found in test results files
    testsuite.print_final_summary(show_minimal=False, show_recommended=False)
    testsuite.print_final_status(result.logger)
    check_iaas.make_reports(context, testsuite)
    return testsuite


if __name__ == '__main__':
    testsuite = main()
    if testsuite.is_any_test_failed():
        sys.exit(1)



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
from typing import Optional, List, Dict

from textwrap import dedent
import yaml

from kubemarine.core import utils, log
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment
from kubemarine.core.group import NodeGroup
from kubemarine.kubernetes import secrets
from kubemarine.plugins.manifest import Processor, EnrichmentFunction, Manifest, Identity

ERROR_CERT_RENEW_NOT_INSTALLED = "Certificates can not be renewed for nginx plugin since it is not installed"


def check_job_for_nginx(cluster: KubernetesCluster) -> None:
    first_control_plane = cluster.nodes['control-plane'].get_first_member()
    version = cluster.inventory['plugins']['nginx-ingress-controller']['version'].replace('v', '.').split('.')

    major_version = int(version[1])
    minor_version = int(version[2])

    check_jobs = first_control_plane.sudo(f"kubectl get jobs -n ingress-nginx")
    if list(check_jobs.values())[0].stderr == "" and major_version >= 1 and minor_version >= 4:
        cluster.log.debug('Delete old jobs for nginx')
        first_control_plane.sudo(f"sudo kubectl delete job --all -n ingress-nginx")
    else:
        cluster.log.debug('There are no jobs to delete')


@enrichment(EnrichmentStage.FULL)
def enrich_inventory(cluster: KubernetesCluster) -> None:
    inventory = cluster.inventory
    if not inventory["plugins"]["nginx-ingress-controller"]["install"]:
        return

    if inventory["plugins"]["nginx-ingress-controller"].get('custom_headers'):
        if not inventory["plugins"]["nginx-ingress-controller"].get('config_map'):
            inventory["plugins"]["nginx-ingress-controller"]['config_map'] = {}
        if not inventory["plugins"]["nginx-ingress-controller"]['config_map'].get('proxy-set-headers'):
            inventory["plugins"]["nginx-ingress-controller"]['config_map']['proxy-set-headers'] = 'ingress-nginx/custom-headers'

    # if user defined resources himself, we should use them as is, instead of merging with our defaults
    raw_controller = cluster.raw_inventory.get("plugins", {}).get("nginx-ingress-controller", {}).get("controller", {})
    if "resources" in raw_controller:
        inventory["plugins"]["nginx-ingress-controller"]["controller"]["resources"] = raw_controller["resources"]
    raw_webhook = cluster.raw_inventory.get("plugins", {}).get("nginx-ingress-controller", {}).get("webhook", {})
    if "resources" in raw_webhook:
        inventory["plugins"]["nginx-ingress-controller"]["webhook"]["resources"] = raw_webhook["resources"]


@enrichment(EnrichmentStage.PROCEDURE, procedures=['cert_renew'])
def cert_renew_enrichment(cluster: KubernetesCluster) -> None:
    # check that renewal is required for nginx
    procedure_nginx_cert = cluster.procedure_inventory.get("nginx-ingress-controller", {})
    if not procedure_nginx_cert:
        return

    # update certificates in inventory
    cluster.inventory.setdefault("plugins", {}).setdefault("nginx-ingress-controller", {}) \
        .setdefault("controller", {}).setdefault("ssl", {})["default-certificate"] \
        = utils.deepcopy_yaml(procedure_nginx_cert)


@enrichment(EnrichmentStage.PROCEDURE, procedures=['cert_renew'])
def verify_cert_renew(cluster: KubernetesCluster) -> None:
    procedure_nginx_cert = cluster.procedure_inventory.get("nginx-ingress-controller", {})
    # check that renewal is possible
    if procedure_nginx_cert and not cluster.inventory["plugins"]["nginx-ingress-controller"]["install"]:
        raise Exception(ERROR_CERT_RENEW_NOT_INSTALLED)


def manage_custom_certificate(cluster: KubernetesCluster) -> None:
    default_cert = cluster.inventory["plugins"]["nginx-ingress-controller"]["controller"]["ssl"]\
        .get("default-certificate")
    if not default_cert:
        cluster.log.debug("No custom default ingress certificate specified, skipping...")
        return

    secret_name = "default-ingress-cert"
    secret_namespace = "kube-system"

    first_control_plane = cluster.nodes["control-plane"].get_first_member()

    # first, we need to load cert and key files to first control-plane to known locations
    with secrets.create_tls_secret_procedure(first_control_plane):
        first_control_plane.call(put_custom_certificate, default_cert=default_cert)

        # second, we need to validate cert and key using openssl
        first_control_plane.call(verify_certificate_and_key)

        # third, we need to create tls secret under well-known name
        # this certificate is already configured to be used by controller
        first_control_plane.call(secrets.renew_tls_secret,
                                 name=secret_name,
                                 namespace=secret_namespace)
    # fourth, we need to remove base path dir when the procedure is exited


def put_custom_certificate(first_control_plane: NodeGroup, default_cert: dict) -> None:
    if default_cert.get("data"):
        cert = io.StringIO(default_cert["data"]["cert"])
        key = io.StringIO(default_cert["data"]["key"])
    else:
        cert = io.StringIO(utils.read_external(default_cert["paths"]["cert"]))
        key = io.StringIO(utils.read_external(default_cert["paths"]["key"]))

    secrets.put_certificate(first_control_plane, cert, key)


def verify_certificate_and_key(first_control_plane: NodeGroup) -> None:
    if not secrets.verify_certificate(first_control_plane):
        raise Exception("Custom default ingress certificate and key are not compatible!")


class IngressNginxManifestProcessor(Processor):
    def __init__(self, logger: log.VerboseLogger, inventory: dict,
                 original_yaml_path: Optional[str] = None, destination_name: Optional[str] = None) -> None:
        super().__init__(logger, inventory, Identity('nginx-ingress-controller'), original_yaml_path, destination_name)

    def get_known_objects(self) -> List[str]:
        return [
            "Namespace_ingress-nginx",
            "ServiceAccount_ingress-nginx",
            "ServiceAccount_ingress-nginx-admission",
            "Role_ingress-nginx",
            "Role_ingress-nginx-admission",
            "ClusterRole_ingress-nginx",
            "ClusterRole_ingress-nginx-admission",
            "RoleBinding_ingress-nginx",
            "RoleBinding_ingress-nginx-admission",
            "ClusterRoleBinding_ingress-nginx",
            "ClusterRoleBinding_ingress-nginx-admission",
            "ConfigMap_ingress-nginx-controller",
            "Service_ingress-nginx-controller",
            "Service_ingress-nginx-controller-admission",
            "Deployment_ingress-nginx-controller",
            "Job_ingress-nginx-admission-create",
            "Job_ingress-nginx-admission-patch",
            "IngressClass_nginx",
            "NetworkPolicy_ingress-nginx-admission",
            "ValidatingWebhookConfiguration_ingress-nginx-admission",
        ]

    def get_enrichment_functions(self) -> List[EnrichmentFunction]:
        return [
            self.enrich_namespace_ingress_nginx,
            self.enrich_configmap_ingress_nginx_controller,
            self.add_configmap_ingress_nginx_controller,
            self.enrich_service_account_secret,
            self.enrich_service_account,
            self.enrich_deployment_ingress_nginx_controller,
            self.enrich_ingressclass_nginx,
            self.enrich_job_ingress_nginx_admission_create,
            self.enrich_job_ingress_nginx_admission_patch,
            self.enrich_service_ingress_nginx_controller,
        ]

    def get_namespace_to_necessary_pss_profiles(self) -> Dict[str, str]:
        return {'ingress-nginx': 'privileged'}

    def enrich_namespace_ingress_nginx(self, manifest: Manifest) -> None:
        self.assign_default_pss_labels(manifest, 'ingress-nginx')

    def enrich_configmap_ingress_nginx_controller(self, manifest: Manifest) -> None:
        key = "ConfigMap_ingress-nginx-controller"
        config_map = self.inventory['plugins']['nginx-ingress-controller'].get('config_map')
        if config_map:
            source_yaml = manifest.get_obj(key, patch=True)
            data: dict = source_yaml['data']
            data.update(config_map)
            self.log.verbose(f"The {key} has been patched in 'data' "
                             f"with the data from 'plugins.nginx-ingress-controller.config_map'")

    def add_configmap_ingress_nginx_controller(self, manifest: Manifest) -> None:
        custom_headers = self.inventory['plugins']['nginx-ingress-controller'].get('custom_headers')
        if custom_headers:
            custom_headers_cm = dict(CUSTOM_HEADERS_CM)
            custom_headers_cm['data'] = custom_headers
            # Insert custom-headers ConfigMap before ingress-nginx-controller ConfigMap
            ingres_nginx_cm = manifest.key_index("ConfigMap_ingress-nginx-controller")
            self.include(manifest, ingres_nginx_cm, custom_headers_cm)
            self.log.verbose(f"The {manifest.obj_key(custom_headers_cm)} has been patched in 'data' "
                             f"with the data from 'plugins.nginx-ingress-controller.custom_headers'")

    def enrich_service_account_secret(self, manifest: Manifest) -> None:
        new_yaml = yaml.safe_load(service_account_secret)

        service_account_key = "ServiceAccount_ingress-nginx"
        service_account_index = manifest.all_obj_keys().index(service_account_key) \
            if service_account_key in manifest.all_obj_keys() else -1

        self.include(manifest, service_account_index + 1, new_yaml)

    def enrich_service_account(self, manifest: Manifest) -> None:
        key = "ServiceAccount_ingress-nginx"
        source_yaml = manifest.get_obj(key, patch=True)
        source_yaml['automountServiceAccountToken'] = False

    def enrich_deployment_ingress_nginx_controller(self, manifest: Manifest) -> None:
        key = "Deployment_ingress-nginx-controller"
        source_yaml = manifest.get_obj(key, patch=True)
        service_account_name = "ingress-nginx"
        self.enrich_volume_and_volumemount(source_yaml, service_account_name)
        self.log.verbose(f"The {key} has been updated to include the new secret volume and mount.")

        self.enrich_deamonset_ingress_nginx_controller_container(manifest)

        self.enrich_image_for_container(manifest, key,
            plugin_service='controller', container_name='controller', is_init_container=False)

        self.enrich_resources_for_container(manifest, key, container_name='controller', plugin_service="controller")
        self.enrich_node_selector(manifest, key, plugin_service='controller')
        self.enrich_tolerations(manifest, key, plugin_service='controller')

        # DeamonSet spec lacks of strategy. Discard it if present.
        source_yaml['spec'].pop('strategy', None)
        # Patch kind in the last step to avoid sudden key change in log messages
        source_yaml['kind'] = 'DaemonSet'
        self.log.verbose(f"The {key} has been patched in 'kind' with 'DaemonSet'")

    def enrich_deamonset_ingress_nginx_controller_container(self, manifest: Manifest) -> None:
        key = "Deployment_ingress-nginx-controller"

        extra_args: List[str] = [
            '--watch-ingress-without-class=true'
        ]
        ssl_options = self.inventory['plugins']['nginx-ingress-controller']['controller']['ssl']

        if ssl_options['enableSslPassthrough']:
            extra_args.append('--enable-ssl-passthrough')
        if ssl_options.get('default-certificate'):
            extra_args.append('--default-ssl-certificate' + '=' + 'kube-system/default-ingress-cert')

        self.enrich_deamonset_ingress_nginx_controller_container_args(
            manifest, remove_args=['--publish-service','--enable-metrics'], extra_args=extra_args)

        container_pos, container = self.find_container_for_patch(
            manifest, key, container_name='controller', is_init_container=False)

        container['ports'] = self.inventory['plugins']['nginx-ingress-controller']['ports']
        self.log.verbose(f"The {key} has been patched in 'spec.template.spec.containers.[{container_pos}].ports' "
                         f"with the data from 'plugins.nginx-ingress-controller.ports'")

    def enrich_deamonset_ingress_nginx_controller_container_args(self, manifest: Manifest,
                                                                 *,
                                                                 remove_args: List[str],
                                                                 extra_args: List[str]) -> None:
        key = "Deployment_ingress-nginx-controller"
        self.enrich_args_for_container(manifest, key,
                                       plugin_service='controller', container_name='controller',
                                       remove_args=remove_args,
                                       extra_args=extra_args)

    def enrich_ingressclass_nginx(self, manifest: Manifest) -> None:
        key = "IngressClass_nginx"
        source_yaml = manifest.get_obj(key, patch=True)
        source_yaml['metadata'].setdefault('annotations', {})['ingressclass.kubernetes.io/is-default-class'] = 'true'
        self.log.verbose(f"The {key} has been patched in 'metadata.annotations' "
                         f"with 'ingressclass.kubernetes.io/is-default-class: true'")

    def enrich_job_ingress_nginx_admission_create(self, manifest: Manifest) -> None:
        key = "Job_ingress-nginx-admission-create"
        self.enrich_image_for_container(manifest, key,
            plugin_service='webhook', container_name='create', is_init_container=False)

        self.enrich_resources_for_container(manifest, key, container_name='create', plugin_service="webhook")

    def enrich_job_ingress_nginx_admission_patch(self, manifest: Manifest) -> None:
        key = "Job_ingress-nginx-admission-patch"
        self.enrich_image_for_container(manifest, key,
            plugin_service='webhook', container_name='patch', is_init_container=False)

        self.enrich_resources_for_container(manifest, key, container_name='patch', plugin_service="webhook")

    def enrich_service_ingress_nginx_controller(self, manifest: Manifest) -> None:
        # The method needs some rework in case of dual stack support
        key = "Service_ingress-nginx-controller"
        ip = self.inventory['services']['kubeadm']['networking']['serviceSubnet'].split('/')[0]
        if utils.isipv(ip, [6]):
            source_yaml = manifest.get_obj(key, patch=True)
            source_yaml['spec']['ipFamilies'] = ['IPv6']
            self.log.verbose(f"The {key} has been patched in 'spec.ipFamilies' with 'IPv6'")


CUSTOM_HEADERS_CM = {
    "apiVersion": "v1",
    "kind": "ConfigMap",
    "metadata": {
        "name": "custom-headers",
        "namespace": "ingress-nginx"
    }
}

service_account_secret = dedent("""\
    apiVersion: v1
    kind: Secret
    metadata:
      name: ingress-nginx-token
      namespace: ingress-nginx
      annotations:
        kubernetes.io/service-account.name: ingress-nginx
    type: kubernetes.io/service-account-token  
""")


#!/usr/bin/env python3
# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
import time
import types
from collections import OrderedDict
from typing import Dict, List

import ruamel.yaml.resolver
import yaml

from kubemarine import procedures as proc

# pylint: disable=bad-builtin

# Don't remove this line. The idna encoding
# is used by getaddrinfo when dealing with unicode hostnames,
# and in some cases, there appears to be a race condition
# where threads will get a LookupError on getaddrinfo() saying
# that the encoding doesn't exist.  Using the idna encoding before
# running any kubemarine code (and any threads it may create) ensures that
# the encodings.idna is imported and registered in the codecs registry,
# which will stop the LookupErrors from happening.
# See: https://bugs.python.org/issue29288
u''.encode('idna')  # pylint: disable=redundant-u-string-prefix

# Take pattern to resolve float used by ruamel instead of that is used by pyyaml.
# https://sourceforge.net/p/ruamel-yaml/code/ci/0.17.21/tree/resolver.py#l43
# Initially introduced for keepalived that generates random password string.
# The generated string might be also a valid exponential float, and should be dumped with quotes.
# See also:
# https://stackoverflow.com/questions/30458977/yaml-loads-5e-6-as-string-and-not-a-number
# https://github.com/yaml/pyyaml/issues/173
for ir in ruamel.yaml.resolver.implicit_resolvers:
    # YAML 1.1 and float implicit resolver
    if (1, 1) in ir[0] and 'tag:yaml.org,2002:float' in ir[1]:
        float_patched_resolver = (ir[1], ir[2], ir[3])
        # Globally change behaviour of yaml.safe_load and yaml.dump
        yaml.Dumper.add_implicit_resolver(*float_patched_resolver)  # type: ignore[no-untyped-call]
        yaml.SafeLoader.add_implicit_resolver(*float_patched_resolver)  # type: ignore[no-untyped-call]
        break


procedures = OrderedDict({
    'install': {
        'description': "Install a cluster from scratch",
        'group': 'installation'
    },
    'migrate_kubemarine': {
        'description': "Automatically perform migration to update the environment for the "
                       "current version of Kubemarine",
        'group': 'maintenance'
    },
    'upgrade': {
        'description': "Automatically upgrade the entire Kubernetes cluster to a new version",
        'group': 'maintenance'
    },
    'backup': {
        'description': "Backup Kubernetes resources and nodes content to backup file",
        'group': 'maintenance'
    },
    'restore': {
        'description': "Restore Kubernetes resources and nodes content from backup file",
        'group': 'maintenance'
    },
    'add_node': {
        'description': "Add new nodes to an existing cluster",
        'group': 'maintenance'
    },
    'remove_node': {
        'description': "Remove existing nodes from cluster",
        'group': 'maintenance'
    },
    'reconfigure': {
        'description': "Reconfigure managed Kubernetes cluster",
        'group': 'maintenance'
    },
    'manage_pss': {
        'description': "Manage PSS on Kubernetes cluster",
        'group': 'maintenance'
    },
    'cert_renew': {
        'description': "Renew certificates on Kubernetes cluster",
        'group': 'maintenance'
    },
    'reboot': {
        'description': "Reboot Kubernetes nodes",
        'group': 'maintenance'
    },
    'check_iaas': {
        'description': "Check environment for compliance with IAAS requirements",
        'group': 'checks'
    },
    'check_paas': {
        'description': "Check environment for compliance with PAAS requirements",
        'group': 'checks'
    },
    'version': {
        'description': "Print current release version",
        'group': 'other'
    },
    'config': {
        'description': "Print the supported k8s versions with the respective configurations of third-parties",
        'group': 'other'
    },
    'do': {
        'description': "Execute shell command on cluster nodes",
        'group': 'other'
    },
    'selftest': {
        'description': "Test internal imports and resources presence",
        'group': 'other'
    },
})


def main() -> None:
    arguments = sys.argv[1:]

    if len(arguments) > 0:
        if arguments[0] == 'selftest':
            return selftest()
        elif arguments[0] == 'version':
            return version()

    if len(arguments) < 1 or arguments[0] not in procedures.keys():
        descriptions_print_list = []
        max_module_name_size = len(max(procedures.keys(), key=len))

        items_description_by_groups: Dict[str, List[str]] = {}

        for module_name, module in procedures.items():
            items_description_by_groups.setdefault(module['group'], [])\
                .append('  %s%s  %s' % (module_name, ' ' * (max_module_name_size - len(module_name)), module['description']))

        previous_group = None
        for group, descriptions in items_description_by_groups.items():
            if group != previous_group:
                descriptions_print_list.append('\n%s:' % group.upper())
                previous_group = group
            for description in descriptions:
                descriptions_print_list.append(' ' + description)
        print('''The following procedures available:
%s

Usage: kubemarine <procedure> <arguments>
''' % '\n'.join(descriptions_print_list))
        sys.exit(1)

    result = proc.import_procedure(arguments[0]).main(arguments[1:])
    if result is not None:
        from kubemarine.testsuite import TestSuite
        if isinstance(result, TestSuite) and result.is_any_test_failed():
            sys.exit(1)


def version() -> None:
    from kubemarine.core import utils

    print('Kubemarine %s' % utils.get_version())


def selftest() -> None:
    print("Running selftest")

    time_start = int(round(time.time() * 1000))

    for procedure in procedures:
        print("\nImporting %s..." % procedure)

        if procedure in ['version', 'selftest']:
            continue

        module = proc.import_procedure(procedure)
        imports = []

        for attr in dir(module):
            if isinstance(getattr(module, attr), types.ModuleType):
                imports.append(attr)

        print("%s has %s imports" % (procedure, len(imports)))

        if isinstance(module, proc.TasksProcedure) and not module.tasks:
            raise Exception("Tasks are empty in %s" % procedure)

        print("%s OK" % procedure)

        del module
        proc.deimport_procedure(procedure)

    print("\nTrying fake cluster...")

    from kubemarine import demo

    demo.new_cluster(demo.generate_inventory(**demo.FULLHA))

    print('\nValidating patch duplicates ...')

    from kubemarine.procedures import migrate_kubemarine
    patches = migrate_kubemarine.load_patches()
    patch_ids = [patch.identifier for patch in patches]
    unique = set()
    for p_id in patch_ids:
        if p_id in unique:
            raise Exception(f'Patches identifier {p_id!r} is duplicated')
        unique.add(p_id)

    print("Finished")

    time_end = int(round(time.time() * 1000))
    print("\nElapsed: %sms\n" % (time_end-time_start))


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import OrderedDict
from typing import List

from kubemarine.core import flow
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.procedures import install
from kubemarine import system


def reboot(cluster: KubernetesCluster) -> None:
    if cluster.context.get('initial_procedure') != 'reboot':
        raise ImportError('Invalid reboot.py usage, please use system.reboot_nodes')

    if not cluster.procedure_inventory.get("nodes"):
        cluster.log.verbose('No nodes defined in procedure: all nodes will be rebooted')
    else:
        cluster.log.verbose('There are nodes defined in procedure: only defined will be rebooted')

    nodes = []

    cluster.log.verbose('The following nodes will be rebooted:')
    for node in cluster.procedure_inventory.get("nodes", cluster.inventory['nodes']):
        nodes.append(node['name'])
        cluster.log.verbose('  - ' + node['name'])

    system.reboot_group(cluster.make_group_from_nodes(nodes),
                        try_graceful=cluster.procedure_inventory.get("graceful_reboot"))


tasks = OrderedDict({
    "reboot": reboot,
    "overview": install.overview,
})


class RebootAction(flow.TasksAction):
    def __init__(self) -> None:
        super().__init__('reboot', tasks)


def create_context(cli_arguments: List[str] = None) -> dict:
    cli_help = '''
    Script for Kubernetes nodes graceful rebooting.

    How to use:

    '''

    parser = flow.new_procedure_parser(cli_help, optional_config=True, tasks=tasks)

    context = flow.create_context(parser, cli_arguments, procedure='reboot')
    return context


def main(cli_arguments: List[str] = None) -> None:
    context = create_context(cli_arguments)
    flow.ActionsFlow([RebootAction()]).run_flow(context)


if __name__ == '__main__':
    main()



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
import urllib.request
import urllib.error
from collections.abc import Hashable
from textwrap import dedent
from typing import List, Dict, Union, cast

import jsonschema
import referencing
import referencing.exceptions
import referencing.retrieval
from ordered_set import OrderedSet

from kubemarine.core import utils, log, errors
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment


@enrichment(EnrichmentStage.LIGHT)
def verify_connections(cluster: KubernetesCluster) -> None:
    inventory = utils.convert_native_yaml(cluster.inventory)
    # Run restricted validation of sections that participate in the enrichment of connections.
    # This is primarily needed for `do`, and `migrate_kubemarine` procedures.
    _verify_inventory_by_schema(cluster, inventory, 'internal/connections')
    context = cluster.context
    args: dict = context['execution_arguments']
    procedure = context["initial_procedure"]
    if args.get('procedure_config') and procedure in ('add_node', 'remove_node'):
        # Run full validation for add_node/remove_node
        # It is currently not necessary to restrict validation to connections only as for inventory.
        procedure_inventory = utils.convert_native_yaml(cluster.procedure_inventory)
        _verify_inventory_by_schema(cluster, procedure_inventory, context["initial_procedure"])


@enrichment(EnrichmentStage.FULL)
def verify_inventory(cluster: KubernetesCluster) -> None:
    inventory = utils.convert_native_yaml(cluster.inventory)
    _verify_inventory_by_schema(cluster, inventory, 'cluster')
    context = cluster.context
    args: dict = context['execution_arguments']
    # Skip validation if procedure inventory is optional and not provided
    if args.get('procedure_config'):
        procedure_inventory = utils.convert_native_yaml(cluster.procedure_inventory)
        _verify_inventory_by_schema(cluster, procedure_inventory, context["initial_procedure"])


def _verify_inventory_by_schema(cluster: KubernetesCluster, inventory: dict, schema_name: str) -> None:
    for_procedure = ("" if schema_name == 'cluster'
                     else " for connections" if schema_name == 'internal/connections'
                     else f" for procedure '{schema_name}'")

    root_schema_resource = f'resources/schemas/{schema_name}.json'
    root_schema_path = utils.get_internal_resource_path(root_schema_resource)
    root_schema = pathlib.Path(root_schema_path)
    if not root_schema.exists():
        raise Exception(f"Failed to find schema to validate the inventory file{for_procedure}.")

    root_schema_uri = root_schema.as_uri()

    registry = referencing.Registry(retrieve=retrieve_uri_filesystem)  # type: ignore[call-arg, var-annotated]
    validator = jsonschema.Draft7Validator({"$ref": root_schema_uri}, registry=registry)

    errs = list(validator.iter_errors(inventory))
    if not errs:
        return

    context = cluster.context
    logger = cluster.log
    logger.error(f"Inventory file{for_procedure} is failed to be validated against the schema.")

    errs = _resolve_errors(errs)
    for err in errs:
        logger.verbose("------------------------------------------")
        logger.verbose(_verbose_msg(validator, err))

    debug_filepath = log.get_dump_debug_filepath(context)
    if debug_filepath:
        detailed_msg = f"See detailed message in {debug_filepath}."
    else:
        detailed_msg = "Enable verbose logs to see details."

    version = utils.get_version()
    public_schema = f"https://raw.githubusercontent.com/Netcracker/KubeMarine/{version}/kubemarine/{root_schema_resource}"
    hint = f"Inventory file{for_procedure} has incorrect format. {detailed_msg}\n" \
           f"To validate the file manually, you can use JSON schema {root_schema_uri}\n" \
           f"or its public alternative {public_schema}\n"

    msg = _error_msg(validator, errs[0])
    raise errors.FailException(msg, hint=hint)


@referencing.retrieval.to_cached_resource()  # type: ignore[arg-type]
def retrieve_uri_filesystem(uri: str) -> str:
    try:
        with urllib.request.urlopen(uri) as url:
            content: str = url.read().decode("utf-8")
            return content
    except urllib.error.URLError:
        raise referencing.exceptions.NoSuchResource(ref=uri) from None  # type: ignore[call-arg]


def _resolve_errors(errs: List[jsonschema.ValidationError]) -> List[jsonschema.ValidationError]:
    key = _extended_relevance
    for error in errs:
        _unnest_errors(error)

    errs.sort(key=key, reverse=True)
    outer_sorted = []
    for error in errs:
        outer_sorted.extend(_descend_errors(error))

    return outer_sorted


def _unnest_errors(error: jsonschema.ValidationError) -> None:
    if not error.context:
        return

    context: List[jsonschema.ValidationError] = error.context
    errors_by_subschema: Dict[Union[str, int], List[jsonschema.ValidationError]] = {}
    for child in context:
        _unnest_errors(child)
        errors_by_subschema.setdefault(child.schema_path[0], []).append(child)

    # jsonschema might show not very friendly messages for anyOf / oneOf.
    # In case of failed type or enum validation in each subschema, we can unnest the error to the higher level.
    # Note that possible remaining subschema errors will be discarded,
    # but it is acceptable for our current schemas implementation.
    subschemas_errors = list(errors_by_subschema.values())
    _unnest_type_subschema_errors(error, subschemas_errors)
    _unnest_enum_subschema_errors(error, subschemas_errors)
    _unnest_required_subschema_errors(error, subschemas_errors)


def _descend_errors(error: jsonschema.ValidationError) -> List[jsonschema.ValidationError]:
    if not error.context:
        return [error]

    # Here can be anyOf or oneOf with all not valid subschemas.
    # The idea is taken from jsonschema.exceptions.best_match,
    # with improved heuristic algorithm of choosing of the best subschema match.
    # 1. group errors by subschema
    # 2. find error with the max relevance in each group
    # 3. choose group with the min relevance of the error found in the previous step
    # 4. take the group instead of the parent error, sort by relevance reverse
    # 5. recurse if necessary

    context: List[jsonschema.ValidationError] = error.context
    errors_by_subschema: Dict[Union[str, int], List[jsonschema.ValidationError]] = {}
    for child in context:
        errors_by_subschema.setdefault(child.schema_path[0], []).append(child)

    key = _extended_relevance

    for errors in errors_by_subschema.values():
        errors.sort(key=key, reverse=True)

    smallest_subschema_errors = sorted(errors_by_subschema.values(), key=lambda errors: key(errors[0]))

    # We don't recurse if two the smallest errors have the same relevance (i.e. if min == max == all).
    # This place is a candidate of future developing of better heuristic.
    if len(smallest_subschema_errors) >= 2 and key(smallest_subschema_errors[0][0]) == key(smallest_subschema_errors[1][0]):
        return [error]

    unnested_errors = []
    for err in smallest_subschema_errors[0]:
        unnested_errors.extend(_descend_errors(err))

    return unnested_errors


def _unnest_type_subschema_errors(error: jsonschema.ValidationError,
                                  subschemas_errors: List[List[jsonschema.ValidationError]]) -> None:
    if not error.context:
        return

    expected_types = OrderedSet[str]()
    for errs in subschemas_errors:
        for child in errs:
            if _validated_by(child, "type") and len(child.relative_path) == 0 and list(child.schema_path)[1:] == ["type"]:
                value = child.validator_value
                expected_types.update([value] if isinstance(value, str) else value)
                break
        else:  # not found error with "type" validation failed for root instance.
            break
    else:  # not found subschema not containing the necessary error, i.e. all subschemas have necessary error
        reprs = ", ".join(repr(type) for type in expected_types)
        for child in error.context:
            child.parent = None
        error.context = []
        error.validator = "type"  # type: ignore[assignment]
        error.validator_value = list(expected_types)
        error.schema_path[-1] = "type"
        error.message = f"{error.instance!r} is not of type {reprs}"
        subschemas_errors.clear()


def _unnest_enum_subschema_errors(error: jsonschema.ValidationError,
                                  subschemas_errors: List[List[jsonschema.ValidationError]]) -> None:
    if not error.context:
        return

    expected_elems = OrderedSet[str]()
    for errs in subschemas_errors:
        for child in errs:
            if _validated_by(child, "enum") and len(child.relative_path) == 0 and list(child.schema_path)[1:] == ["enum"]:
                expected_elems.update(child.validator_value)
                break
        else:  # not found error with "enum" validation failed for root instance.
            break
    else:  # not found subschema not containing the necessary error, i.e. all subschemas have necessary error
        for child in error.context:
            child.parent = None
        error.context = []
        error.validator = "enum"  # type: ignore[assignment]
        error.validator_value = list(expected_elems)
        error.schema_path[-1] = "enum"
        error.message = f"{error.instance!r} is not one of {list(expected_elems)!r}"
        subschemas_errors.clear()


def _unnest_required_subschema_errors(error: jsonschema.ValidationError,
                                      subschemas_errors: List[List[jsonschema.ValidationError]]) -> None:
    if not error.context:
        return

    required_sets: List[List[str]] = []
    for errs in subschemas_errors:
        if len(errs) > 1:
            # Support only the case when all subschema failed with the only 'required' validation.
            # This is a common pattern if "at least one property from the set is required" requirement
            # is moved to the separate "allOf (anyOf (required prop1, required prop2)) subschema.
            break
        for child in errs:
            if (_validated_by(child, "required") and len(child.relative_path) == 0
                    and list(child.schema_path)[1:] == ["required"] and len(child.validator_value) > 0
                    and child.validator_value != ['<<']):
                if isinstance(child.validator_value[0], str):
                    required_sets.append(child.validator_value)
                else:
                    required_sets.extend(child.validator_value)
                break
        else:  # not found error with "required" validation failed for root instance.
            break
    # not found subschema not containing the only necessary error,
    # i.e. all subschemas have the only necessary error
    else:
        initial_validator = cast(str, error.validator)

        for child in error.context:
            child.parent = None

        error.context = []
        error.validator = "required"  # type: ignore[assignment]
        # This breaks original typing.
        # Access to the validator_value should always be guarded with extra isinstance() checks.
        error.validator_value = required_sets
        error.schema_path[-1] = "required"
        error.message = (f"{'One' if initial_validator == 'oneOf' else 'At least one'} "
                         f"of the following property sets is required: {', '.join(map(str, required_sets))}")
        subschemas_errors.clear()


def _extended_relevance(error: jsonschema.ValidationError) -> tuple:
    # jsonschema v4.22.0 introduced an improvement of default relevance for `anyOf` / `oneOf`.
    # It solves the problem very partially, and we have own algorithm more suitable for our schemes.
    # Pin algorithm of jsonschema.exceptions.relevance for jsonschema < 4.22.0
    is_week = any(_validated_by(error, validator)
                  for validator in jsonschema.exceptions.WEAK_MATCHES)
    matches_type = error._matches_type()  # type: ignore[attr-defined] # pylint: disable=protected-access
    relevance_value: tuple = (      # prefer errors which are ...
        -len(error.path),           # 'deeper' and thereby more specific
        not is_week,                # for a non-low-priority keyword
        not matches_type,           # at least match the instance's type
    )                               # otherwise we'll treat them the same

    # The extended relevance function is intended to improve heuristic for oneOf|anyOf,
    # when it is necessary to choose the most suitable branch.
    if error.parent is None:
        return relevance_value

    relevance_value = _apply_property_names_heuristic(error, relevance_value)
    relevance_value = _apply_list_merging_strong_heuristic(error, relevance_value)
    relevance_value = _apply_required_and_optional_properties_heuristic(error, relevance_value)

    return relevance_value


def _apply_property_names_heuristic(error: jsonschema.ValidationError, relevance_value: tuple) -> tuple:
    # jsonschema has type matching heuristic but it works bad for "propertyNames".
    # "propertyNames" does not introduce new path element and effectively verifies the object holding the property.
    # But an attempt to match the type happens for the "propertyNames" subschema.
    # The following heuristic resolves the simplest case of oneOf|anyOf(something, schema with "propertyNames" section)

    if len(error.relative_path) == 0 and "propertyNames" in error.schema_path:
        # "propertyNames" is validated only if the instance is "object".
        # See jsonschema._validators.propertyNames. So type is always matched.
        type_matched = True
        mutate_relevance = list(relevance_value)
        mutate_relevance[2] = not type_matched
        return tuple(mutate_relevance)

    return relevance_value


def _apply_list_merging_strong_heuristic(error: jsonschema.ValidationError, relevance_value: tuple) -> tuple:
    # Other conditions being equal, the error for list merging strategy has greater relevance,
    # because if the user specified '<<', we consider that he/she intends to use this advanced feature.

    is_list_merging = len(error.relative_path) == 0 and isinstance(error.schema, dict) \
                      and "properties" in error.schema and error.schema["properties"].keys() == {"<<"}

    return _append_relevance_element(relevance_value, is_list_merging)


def _apply_required_and_optional_properties_heuristic(error: jsonschema.ValidationError,
                                                      relevance_value: tuple) -> tuple:
    # Resolve oneOf|anyOf(object, object) ambiguity.
    # Currently only 'registry' section has such schema.
    # The chosen priority is:
    # 1. If all required properties are present, then the error has lower relevance.
    # 2. Otherwise, calculate proportion of matched properties.

    # By default, consider all required properties are present, and no properties are matched
    coeff: list = [False, 0]
    if len(error.relative_path) != 0 or not isinstance(error.schema, dict) or not isinstance(error.instance, dict):
        return _append_relevance_element(relevance_value, tuple(coeff))

    coeff[0] = not all(p in error.instance for p in error.schema.get("required", []))

    if len(error.schema.get("properties", {})) > 0:
        inverse_weight = len(error.schema["properties"])
        matched = set(error.instance.keys()).intersection(set(error.schema["properties"].keys()))
        # negative coefficient because more matched properties - more suitable subschema - less relevant error
        coeff[1] = - len(matched) / inverse_weight

    return _append_relevance_element(relevance_value, tuple(coeff))


def _append_relevance_element(relevance_value: tuple, elem: Hashable) -> tuple:
    mutate_relevance = list(relevance_value)
    mutate_relevance.append(elem)
    return tuple(mutate_relevance)


def _error_msg(validator: jsonschema.Draft7Validator, error: jsonschema.ValidationError) -> str:
    return dedent(
        f"""\
        {_friendly_msg(validator, error)}
        On inventory{utils.pretty_path(error.absolute_path)}
        """.rstrip()
    )


def _verbose_msg(validator: jsonschema.Draft7Validator, error: jsonschema.ValidationError) -> str:
    schema_path = utils.pretty_path(list(error.absolute_schema_path)[:-1])
    return dedent(
        f"""\
        {_friendly_msg(validator, error)}
        
        Failed validating {error.validator!r} in {schema_path}
        On inventory{utils.pretty_path(error.absolute_path)}
        """.rstrip()
    )


def _friendly_msg(validator: jsonschema.Draft7Validator, error: jsonschema.ValidationError) -> str:
    # pylint: disable=too-many-return-statements

    if _validated_by(error, 'minProperties'):
        return f"Number of properties is less than the minimum of {error.validator_value!r}. " \
               f"Property names: {list(error.instance)!r}"

    if _validated_by(error, 'maxProperties'):
        return f"Number of properties is greater than the maximum of {error.validator_value!r}. " \
               f"Property names: {list(error.instance)!r}"

    if _validated_by(error, 'minItems'):
        return f"Number of items equal to {len(error.instance)} is less than the minimum of {error.validator_value!r}"

    if _validated_by(error, 'maxItems'):
        return f"Number of items equal to {len(error.instance)} is greater than the maximum of {error.validator_value!r}"

    if _validated_by(error, 'type'):
        value = error.validator_value
        expected_types = [value] if isinstance(value, str) else value
        reprs = ", ".join(repr(type) for type in expected_types)
        try:
            for type_ in validator.TYPE_CHECKER._type_checkers:  # pylint: disable=protected-access
                if validator.is_type(error.instance, type_):  # type: ignore[no-untyped-call]
                    return f"Actual instance type is {type_!r}. Expected: {reprs}."
        except (AttributeError, TypeError, jsonschema.exceptions.UnknownType):
            # In current 3rd-party version the error should not appear, but let's still fall back to default behaviour
            pass

    if (_validated_by(error, 'not') and isinstance(error.validator_value, dict)
            and set(error.validator_value) == {'enum'} and "propertyNames" in error.schema_path):
        return f"Property name {error.instance!r} is unexpected"

    if _validated_by(error, 'enum'):
        if "propertyNames" not in error.schema_path:
            return f"Value should be one of {error.validator_value!r}"
        else:
            return f"Property name {error.instance!r} is not one of {error.validator_value!r}"

    return error.message


def _validated_by(error: jsonschema.ValidationError, expected: str) -> bool:
    return error.validator == expected  # type: ignore[comparison-overlap]



# Copyright 2021-2023 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
from typing import List, Union

from ordered_set import OrderedSet

from kubemarine import kubernetes, sysctl, system
from kubemarine.core import flow
from kubemarine.core.cluster import KubernetesCluster


def system_prepare_system_sysctl(cluster: KubernetesCluster) -> None:
    group = cluster.nodes['all']

    # Even if services.sysctl is not supplied, the effective configuration may change
    # because it may depend on other services (kubelet config and patches).
    changes_detected = any(
        cluster.previous_nodes_inventory[host]['services']['sysctl'] != cluster.nodes_inventory[host]['services']['sysctl']
        for host in cluster.nodes['all'].get_hosts()
    )

    # Reconfigure kernel parameters even if empty section is supplied.
    # This allows to reconfigure the parameters based on manual changes in the inventory.
    # Take into account only `services.sysctl`, but not `patches`, because `patches` may reconfigure different services.
    if (cluster.procedure_inventory.get('services', {}).get('sysctl') is not None
            or changes_detected):
        cluster.log.debug(f"Detected changes in kernel parameters")

        # In comparison to installation, do not reboot & verify the parameters.
        # The parameters are closely tied to the other services' settings, e.g. kube-proxy, and may be changed together.
        # After reboot, if sysctl is reconfigured, but other services are not yet,
        # verification may fail, but the final inventory may still be correct after all.
        is_updated = group.call(sysctl.setup_sysctl)
        if is_updated:
            group.call(system.verify_sysctl)
    else:
        cluster.log.debug("No changes detected, skipping.")


def deploy_kubernetes_reconfigure(cluster: KubernetesCluster) -> None:
    changed_components = OrderedSet[str]()
    for component, constants in kubernetes.components.COMPONENTS_CONSTANTS.items():
        for section_names in constants['sections']:
            procedure_section: Union[dict, list, None] = cluster.procedure_inventory
            previous_section = cluster.previous_inventory
            section = cluster.inventory
            for name in section_names:
                if isinstance(procedure_section, dict):
                    procedure_section = procedure_section.get(name)

                previous_section = previous_section[name]
                section = section[name]

            # Even if section is not supplied, the effective configuration may change
            # because it may depend on other services (services.sysctl).
            changed_detected = previous_section != section

            # Consider component as changed even if empty section is supplied.
            # This allows to reconfigure the component based on manual changes in the inventory.
            if (procedure_section is not None
                    or changed_detected):
                changed_components.add(component)

    if changed_components:
        cluster.log.debug(f"Detected changes in components: {', '.join(changed_components)}")
        kubernetes_nodes = cluster.make_group_from_roles(['control-plane', 'worker'])
        kubernetes_nodes.call(kubernetes.components.reconfigure_components, components=list(changed_components))
    else:
        cluster.log.debug("No changes detected, skipping.")


tasks = OrderedDict({
    "prepare": {
        "system": {
            "sysctl": system_prepare_system_sysctl,
        },
    },
    "deploy": {
        "kubernetes": {
            "reconfigure": deploy_kubernetes_reconfigure
        }
    }
})


class ReconfigureAction(flow.TasksAction):
    def __init__(self) -> None:
        super().__init__('reconfigure', tasks, recreate_inventory=True)


def create_context(cli_arguments: List[str] = None) -> dict:
    cli_help = '''
        Script for generic reconfiguring of existing Kubernetes cluster.

        How to use:

        '''

    parser = flow.new_procedure_parser(cli_help, tasks=tasks)
    context = flow.create_context(parser, cli_arguments, procedure="reconfigure")
    return context


def main(cli_arguments: List[str] = None) -> None:
    context = create_context(cli_arguments)
    flow.ActionsFlow([ReconfigureAction()]).run_flow(context)


if __name__ == '__main__':
    main()



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import re
from typing import Dict, List, Optional, Union

import ruamel.yaml
from jinja2 import Template

from kubemarine.core import utils
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment
from kubemarine.core.group import NodeGroup, RunnersGroupResult
from kubemarine.core.yaml_merger import default_merger
from kubemarine.kubernetes import components
from kubemarine.plugins import builtin

admission_template = "./templates/admission.yaml.j2"
admission_dir = "/etc/kubernetes/pki"
admission_path = "%s/admission.yaml" % admission_dir

valid_modes = ['enforce', 'audit', 'warn']
valid_versions_templ = r"^v1\.\d{1,2}$"

ERROR_PSS_BOTH_STATES_DISABLED = ("both 'pod-security' in procedure config and current config are 'disabled'. "
                                  "There is nothing to change")


def is_pod_security_unconditional(cluster: KubernetesCluster) -> bool:
    return components.kubernetes_minor_release_at_least(cluster.inventory, "v1.28")


@enrichment(EnrichmentStage.FULL)
def enrich_inventory(cluster: KubernetesCluster) -> None:
    inventory = cluster.inventory
    # check flags, enforce and logs parameters
    kubernetes_version = inventory["services"]["kubeadm"]["kubernetesVersion"]

    for item, state in cluster.inventory["rbac"]["pss"]["defaults"].items():
        if item.endswith("version"):
            verify_version(item, state, kubernetes_version)

    if not is_security_enabled(inventory):
        return

    # add extraArgs to kube-apiserver config
    extra_args = inventory["services"]["kubeadm"]["apiServer"]["extraArgs"]
    if not is_pod_security_unconditional(cluster):
        enabled_admissions = extra_args.get("feature-gates")
        if enabled_admissions:
            if 'PodSecurity=true' not in enabled_admissions:
                enabled_admissions = "%s,PodSecurity=true" % enabled_admissions
        else:
            enabled_admissions = "PodSecurity=true"

        extra_args["feature-gates"] = enabled_admissions

    extra_args["admission-control-config-file"] = admission_path


def verify_version(owner: str, version: str, kubernetes_version: str) -> None:
    # check Kubernetes version and admission config matching
    if version != "latest":
        result = re.match(valid_versions_templ, version)
        if result is None:
            raise Exception(f"Incorrect {owner} {version!r}, "
                            f"valid version (for example): {utils.minor_version(kubernetes_version)}")
        if utils.minor_version_key(version) > utils.version_key(kubernetes_version)[0:2]:
            raise Exception(f"{owner} must not be higher than Kubernetes version")


def restart_pods_task(cluster: KubernetesCluster) -> None:
    from kubemarine import kubernetes  # pylint: disable=cyclic-import

    if cluster.context.get('initial_procedure') == 'manage_pss':
        # check if pods restart is enabled
        is_restart = cluster.procedure_inventory.get("restart-pods", False)
        if not is_restart:
            cluster.log.debug("'restart-pods' is disabled, pods won't be restarted")
            return

    first_control_plane = cluster.nodes["control-plane"].get_first_member()

    cluster.log.debug("Drain-Uncordon all nodes to restart pods")
    kube_nodes = cluster.nodes["control-plane"].include_group(cluster.nodes["worker"])
    for node in kube_nodes.get_ordered_members_list():
        first_control_plane.sudo(
            kubernetes.prepare_drain_command(cluster, node.get_node_name(), disable_eviction=False),
            hide=False, pty=True)
        first_control_plane.sudo("kubectl uncordon %s" % node.get_node_name(), hide=False)

    cluster.log.debug("Restarting daemon-sets...")
    daemon_sets = ruamel.yaml.YAML().load(list(first_control_plane.sudo("kubectl get ds -A -o yaml").values())[0].stdout)
    for ds in daemon_sets["items"]:
        first_control_plane.sudo("kubectl rollout restart ds %s -n %s" % (ds["metadata"]["name"], ds["metadata"]["namespace"]))

    # we do not know to wait for, only for system pods maybe
    kube_nodes.call(components.wait_for_pods)


def is_security_enabled(inventory: dict) -> bool:
    target_state: str = inventory["rbac"]["pss"]["pod-security"]
    return target_state == "enabled"


@enrichment(EnrichmentStage.PROCEDURE, procedures=['manage_pss'])
def verify_manage_enrichment(cluster: KubernetesCluster) -> None:
    procedure_config = cluster.procedure_inventory["pss"]
    kubernetes_version = cluster.previous_inventory["services"]["kubeadm"]["kubernetesVersion"]

    if (cluster.previous_inventory["rbac"]["pss"]["pod-security"] == "disabled"
            and cluster.inventory["rbac"]["pss"]["pod-security"] == "disabled"):
        raise Exception(ERROR_PSS_BOTH_STATES_DISABLED)

    if "namespaces" in procedure_config:
        for namespace_item in procedure_config["namespaces"]:
            # check if the namespace has its own profiles
            if isinstance(namespace_item, dict):
                profiles = list(namespace_item.values())[0]
                for item in profiles:
                    if item.endswith("version"):
                        verify_version(item, profiles[item], kubernetes_version)
    if "namespaces_defaults" in procedure_config:
        for item in procedure_config["namespaces_defaults"]:
            if item.endswith("version"):
                verify_version(item, procedure_config["namespaces_defaults"][item], kubernetes_version)


def manage_pss(cluster: KubernetesCluster) -> None:
    control_planes = cluster.nodes["control-plane"]

    target_state = cluster.inventory["rbac"]["pss"]["pod-security"]

    # set labels for predefined plugins namespaces and namespaces defined in procedure config
    label_namespace_pss(cluster)

    # copy admission config on control-planes
    copy_pss(control_planes)

    # Admission configuration may change.
    # Force kube-apiserver pod restart, then wait for API server to become available.
    force_restart = True

    # Extra args of API may change, need to reconfigure the API server.
    # See enrich_inventory_pss()
    control_planes.call(components.reconfigure_components,
                        components=['kube-apiserver'], force_restart=force_restart)

    if target_state == 'disabled':
        # erase PSS admission config
        cluster.log.debug("Erase admission configuration... %s" % admission_path)
        control_planes.sudo("rm -f %s" % admission_path, warn=True)


@enrichment(EnrichmentStage.PROCEDURE, procedures=['manage_pss'])
def manage_enrichment(cluster: KubernetesCluster) -> None:
    procedure_config = cluster.procedure_inventory["pss"]

    current_config = cluster.inventory.setdefault("rbac", {}).setdefault("pss", {})

    # merge flags from procedure config and cluster config
    current_config["pod-security"] = procedure_config["pod-security"]
    if "defaults" in procedure_config:
        default_merger.merge(current_config.setdefault("defaults", {}), procedure_config["defaults"])
    if "exemptions" in procedure_config:
        default_merger.merge(current_config.setdefault("exemptions", {}), procedure_config["exemptions"])


def generate_pss(cluster: KubernetesCluster) -> str:
    defaults = cluster.inventory["rbac"]["pss"]["defaults"]
    exemptions = cluster.inventory["rbac"]["pss"]["exemptions"]
    return Template(utils.read_internal(admission_template))\
        .render(defaults=defaults, exemptions=exemptions)


def copy_pss(group: NodeGroup) -> Optional[RunnersGroupResult]:
    if not is_security_enabled(group.cluster.inventory):
        group.cluster.log.debug("Pod security disabled, skipping pod admission installation...")
        return None

    # create admission config from template and cluster.yaml
    admission_config = generate_pss(group.cluster)

    # put admission config on every control-planes
    group.cluster.log.debug(f"Copy admission config to {admission_path}")
    group.put(io.StringIO(admission_config), admission_path, backup=True, sudo=True, mkdir=True)

    return group.sudo(f'ls -la {admission_path}')


def _get_default_labels(profile: str) -> Dict[str, str]:
    return {f"pod-security.kubernetes.io/{k}": v
            for mode in valid_modes
            for k, v in ((mode, profile), (f'{mode}-version', 'latest'))}


def get_labels_to_ensure_profile(inventory: dict, profile: str) -> Dict[str, str]:
    enforce_profile: str = inventory['rbac']['pss']['defaults']['enforce']
    if (enforce_profile == 'restricted' and profile != 'restricted'
            or enforce_profile == 'baseline' and profile == 'privileged'):
        return _get_default_labels(profile)

    return {}


def label_namespace_pss(cluster: KubernetesCluster) -> None:
    security_enabled = is_security_enabled(cluster.inventory)
    first_control_plane = cluster.nodes["control-plane"].get_first_member()
    # set/delete labels on predifined plugins namsespaces
    for ns_name, profile in builtin.get_namespace_to_necessary_pss_profiles(cluster).items():
        target_labels = get_labels_to_ensure_profile(cluster.inventory, profile)
        if security_enabled and target_labels:
            cluster.log.debug(f"Set PSS labels for profile {profile} on namespace {ns_name}")
            command = "kubectl label ns {namespace} {lk}={lv} --overwrite"

        else:  # Pod Security is disabled or default labels are not necessary
            cluster.log.debug(f"Delete PSS labels from namespace {ns_name}")
            command = "kubectl label ns {namespace} {lk}- || true"
            target_labels = _get_default_labels(profile)

        for lk, lv in target_labels.items():
            first_control_plane.sudo(command.format(namespace=ns_name, lk=lk, lv=lv))

    procedure_config = cluster.procedure_inventory["pss"]
    namespaces: List[Union[str, Dict[str, dict]]] = procedure_config.get("namespaces")
    # get the list of namespaces that should be labeled then set/delete labels
    if namespaces:
        default_modes = {}
        # check if procedure config has default values for labels
        namespaces_defaults = procedure_config.get("namespaces_defaults")
        if namespaces_defaults:
            for default_mode in namespaces_defaults:
                default_modes[default_mode] = namespaces_defaults[default_mode]
        for namespace in namespaces:
            # define name of namespace
            if isinstance(namespace, dict):
                ns_name = list(namespace.keys())[0]
            else:
                ns_name = namespace
            if security_enabled:
                if default_modes:
                    # set labels that are set in default section
                    cluster.log.debug(f"Set PSS labels on {ns_name} namespace from defaults")
                    for mode, value in default_modes.items():
                        first_control_plane.sudo(f"kubectl label ns {ns_name} "
                                f"pod-security.kubernetes.io/{mode}={value} --overwrite")
                if isinstance(namespace, dict):
                    # set labels that are set in namespaces section
                    cluster.log.debug(f"Set PSS labels on {ns_name} namespace")
                    for item in list(namespace[ns_name]):
                        first_control_plane.sudo(f"kubectl label ns {ns_name} " 
                                    f"pod-security.kubernetes.io/{item}={namespace[ns_name][item]} --overwrite")
            else:
                # delete labels that are set in default section
                if default_modes:
                    cluster.log.debug(f"Delete PSS labels on {ns_name} namespace from defaults")
                    for mode in default_modes:
                        first_control_plane.sudo(f"kubectl label ns {ns_name} pod-security.kubernetes.io/{mode}-")
                # delete labels that are set in namespaces section
                cluster.log.debug(f"Delete PSS labels on {ns_name} namespace")
                if isinstance(namespace, dict):
                    for item in list(namespace[ns_name]):
                        first_control_plane.sudo(f"kubectl label ns {ns_name} "
                                    f"pod-security.kubernetes.io/{item}-")



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import threading
import time
from contextlib import contextmanager
from typing import Dict, List, Any, Optional, cast, Callable, Tuple, Iterator, IO, Union

import fabric  # type: ignore[import-untyped]
import fabric.transfer  # type: ignore[import-untyped]
import invoke
import paramiko

from kubemarine.core import static

input_sleep = fabric.Remote.input_sleep


class CommandInterrupted(invoke.Failure):
    pass


class RemoteRunner(fabric.Remote):  # type: ignore[misc]
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.KM_connection: Connection = cast(Connection, self.context)
        self.KM_interrupted = False

    def wait(self) -> None:
        while True:
            # Never raise KeyboardInterrupt. Remote result should be eventually returned, or raised.
            # This only sends SIGINT character, does not raise exception, and makes sense only with `pty`.
            # See invoke.Runner._finish() and fabric.Remote.send_interrupt()
            if self.using_pty and self.KM_connection.KM_interrupt_queue.acquire(blocking=False):
                self.KM_interrupted = True
                raise KeyboardInterrupt()

            # This repeats the behaviour of `invoke.Runner.wait()`
            if self.process_is_finished or self.has_dead_threads:
                break
            time.sleep(input_sleep)

    def generate_result(self, **kwargs: Any) -> fabric.Result:
        result = super().generate_result(**kwargs)

        # invoke converts CRs only on Windows. Let's always convert them to have stable output in memory for parsing.
        if not invoke.terminals.WINDOWS:
            result.stdout = result.stdout.replace("\r\n", "\n").replace("\r", "\n")
            result.stderr = result.stderr.replace("\r\n", "\n").replace("\r", "\n")

        # Raise CommandInterrupted only if pty is used and real SIGINT character is sent.
        if self.KM_interrupted:
            raise CommandInterrupted(result)

        return result


class TransferCallback:
    def __init__(self, connection: 'Connection'):
        self.connection = connection
        self.channel_lock = threading.Lock()
        self.interrupted = False

    def __call__(self, size: int, file_size: int) -> None:
        if self.connection.KM_interrupt_queue.acquire(blocking=False):
            sftp = self.connection.sftp()
            # This opens new SFTP channel on next access to SFTP client using the same connection.
            self.connection._sftp = None

            # We use prefetch for `Connection.get()` that fetches data in background.
            # If just raise KeyboardInterrupt, paramiko will attempt to close the remote file,
            # that will be queued after the prefetch.
            # In contrast, force closing of the SFTP channel does not hang up.
            # This is considered OK, as the remote file is eventually released,
            # and the only expected flaw is the channel can no longer be used.
            with self.channel_lock:
                self.interrupted = True
                sftp.close()

            raise KeyboardInterrupt()


class SFTPClient(paramiko.SFTPClient):
    def __init__(self, sock: paramiko.Channel):
        super().__init__(sock)
        self.KM_callback: Optional[TransferCallback] = None

    def getfo(self, remotepath: Union[bytes, str], fl: IO[bytes],
              callback: Callable[[int, int], object] = None, prefetch: bool = True,
              max_concurrent_prefetch_requests: int = None) -> int:
        return super().getfo(remotepath, fl,
                             callback=self.KM_callback, prefetch=True,
                             max_concurrent_prefetch_requests=max_concurrent_prefetch_requests)

    def putfo(self, fl: IO[bytes], remotepath: Union[bytes, str],
              file_size: int = 0, callback: Callable[[int, int], object] = None,
              confirm: bool = True) -> paramiko.SFTPAttributes:
        return super().putfo(fl, remotepath,
                             file_size=file_size, callback=self.KM_callback, confirm=confirm)

    def open(self, filename: Union[bytes, str], mode: str = "r", bufsize: int = -1) -> paramiko.SFTPFile:
        sftp_file = super().open(filename, mode, bufsize)

        # Paramiko starts daemon thread that prefetches the file.
        # If channel is closed from TransferCallback, the thread fails with an exception.
        # This monkey-patching is only needed to stop it gracefully.
        def _prefetch_thread(file: paramiko.SFTPFile, chunks: List[Tuple[int, int]],
                             max_concurrent_requests: int = None) -> None:
            for offset, length in chunks:
                callback = self.KM_callback
                if callback is None:
                    return
                with callback.channel_lock:
                    if callback.interrupted:
                        return

                    paramiko.SFTPFile._prefetch_thread(  # type: ignore[attr-defined] # pylint: disable=protected-access
                        file, [(offset, length)], max_concurrent_requests)

        # pylint: disable-next=protected-access,assignment-from-none,no-value-for-parameter
        sftp_file._prefetch_thread = _prefetch_thread.__get__(sftp_file, paramiko.SFTPFile)  # type: ignore[attr-defined]
        return sftp_file


class Connection(fabric.Connection):  # type: ignore[misc]
    def __init__(self, host: str, **kwargs: Any):
        super().__init__(host, **kwargs)
        self._sftp: Optional[SFTPClient] = None
        self.KM_interrupt_queue = threading.Semaphore(0)

    def __setattr__(self, key: str, value: Any) -> None:
        # fabric Connection has special handling of this method. Call default behaviour for custom attributes.
        if key in ('_sftp', 'KM_interrupt_queue'):
            return object.__setattr__(self, key, value)
        super().__setattr__(key, value)

    @fabric.connection.opens  # type: ignore[misc]
    def sftp(self) -> SFTPClient:
        if self._sftp is None:
            self._sftp = cast(SFTPClient, SFTPClient.from_transport(self.client.get_transport()))
        return self._sftp

    def get(self, *args: Any, **kwargs: Any) -> fabric.transfer.Result:
        with self.KM_transfer():
            return super().get(*args, **kwargs)

    def put(self, *args: Any, **kwargs: Any) -> fabric.transfer.Result:
        with self.KM_transfer():
            return super().put(*args, **kwargs)

    @contextmanager
    def KM_transfer(self) -> Iterator[None]:
        sftp = self.sftp()
        try:
            sftp.KM_callback = TransferCallback(self)
            yield
        finally:
            sftp.KM_callback = None

    def KM_start(self) -> None:
        self.KM_interrupt_queue = threading.Semaphore(0)

    def KM_interrupt(self) -> None:
        self.KM_interrupt_queue.release()


class ConnectionPool:
    def __init__(self, nodes: Dict[str, dict], gateway_nodes: Dict[str, dict], hosts: List[str]):
        self._nodes = nodes
        self._gateway_nodes = gateway_nodes
        self._connections = {ip: self._create_connection(ip) for ip in hosts}

    def get_node(self, ip: str) -> dict:
        node = self._nodes.get(ip)
        if node is None:
            raise Exception("Failed to find suitable node to connect to by address %s" % ip)

        return node

    def get_connection(self, ip: str) -> Connection:
        conn = self._connections.get(ip)
        if conn is None:
            raise Exception(f'Connection for {ip} is not registered')

        return conn

    def close(self) -> None:
        for conn in self._connections.values():
            conn.close()

        self._connections.clear()

    def _create_connection_from_details(self, ip: str, conn_details: dict,
                                        gateway: fabric.connection.Connection = None,
                                        inline_ssh_env: bool = True) -> Connection:

        connection_defaults = static.GLOBALS['connection']['defaults']
        connect_kwargs = {}
        if conn_details.get('keyfile'):
            connect_kwargs['key_filename'] = os.path.expanduser(conn_details['keyfile'])
        elif conn_details.get('password'):
            connect_kwargs['password'] = conn_details.get('password')

        # connect_timeout is for TCP connect, while channel_timeout is for SSH machinery.
        # Although they have different nature, there is no request to separate them for now.
        # channel_timeout can also be "worked around" by reconnecting the whole connection (including TCP connect)
        connect_timeout = conn_details.get('connection_timeout', connection_defaults['timeout'])
        connect_kwargs['channel_timeout'] = connect_timeout

        cfg = fabric.Config(overrides={
            'run': {'encoding': "utf-8"},
            'runners': {'remote': RemoteRunner},
        })
        return Connection(
            ip,
            user=conn_details.get('username', connection_defaults['username']),
            gateway=gateway,
            port=conn_details.get('connection_port', connection_defaults['port']),
            config=cfg,
            connect_timeout=connect_timeout,
            connect_kwargs=connect_kwargs,
            inline_ssh_env=inline_ssh_env
        )

    def _create_connection(self, ip: str) -> Connection:
        node = self.get_node(ip)

        if node.get('keyfile') is None and node.get('password') is None:
            raise Exception('There is neither keyfile nor password specified in configfile for node \'%s\'' % node['name'])

        gateway = None
        if 'gateway' in node:
            gateway = self._get_gateway_node_connection(node['gateway'])

        return self._create_connection_from_details(ip, node, gateway=gateway)

    def _get_gateway_node_connection(self, name: str) -> Connection:
        # Create new connection instance each time even if it is the same gateway node.
        # This is necessary to not share the same gateway connection instance in multiple threads

        gateway = self._gateway_nodes.get(name)
        if gateway is None:
            raise Exception('Requested gateway \'%s\' is not found in configfile' % name)

        return self._create_connection_from_details(gateway["address"], gateway, inline_ssh_env=False)
